--[[
Interface: 1.7.1.0 b10490

Copyright (C) GtX (Andy), 2020

Author: GtX | Andy
Date: 21.08.2020
Version: 1.0.0.0

Contact:
https://forum.giants-software.com
https://github.com/GtX-Andy

Important:
Free for use in other mods - no permission needed.
No changes are to be made to this script without permission from GtX | Andy

Frei verwendbar - keine erlaubnis nötig
An diesem Skript dürfen ohne Genehmigung von GtX | Andy keine Änderungen vorgenommen werden

Required:
SellingStationTrainLoader.lua
SellingStationTrainPlaceable.lua
]]


SellingStationTrain = {}

SellingStationTrain.DIRECTION_PARKED = 0

SellingStationTrain.DIRECTION_FOWARD = 1
SellingStationTrain.DIRECTION_FOWARD_WAITING = 2

SellingStationTrain.DIRECTION_REVERSE = -1
SellingStationTrain.DIRECTION_REVERSE_WAITING = -2

SellingStationTrain.NAME_TO_DIRECTION = {
    ["waitingReverse"] = -2,
    ["reverse"] = -1,
    ["parked"] = 0,
    ["forward"] = 1,
    ["waitingForward"] = 2
}

SellingStationTrain.DIRECTION_P3_TO_NAME = {
    "waitingReverse",
    "reverse",
    "parked",
    "forward",
    "waitingForward"
}

SellingStationTrain.DEFAULT_DIRECTIONS = "waitingReverse reverse forward waitingForward"

local SellingStationTrain_mt = Class(SellingStationTrain, Object)
InitObjectClass(SellingStationTrain, "SellingStationTrain")

function SellingStationTrain:new(isServer, isClient, customMt)
    local self = Object:new(isServer, isClient, customMt or SellingStationTrain_mt)

    self.lights = {}
    self.samples = {}
    self.fillPlanes = {}

    self.fillLevels = {}
    self.lastAdded = {}
    self.soldPrices = {}

    self.needsUpdate = false

    self.startingUserName = "Unknown"
    self.startingUserId = 0

    self.timeSinceFillTypesAdded = -1

    self.splineTime = 0
    self.splineTimeSent = 0

    self.lastSpeed = 0
    self.lastMovingSpeed = 0

    self.travelDirection = 0
    self.waitTime = 0

    self.brakingSpeed = 0
    self.startBrakeSpeed = 0

    self.surfaceSounds = {}
    self.locomotiveStartTime = -1

    self.trainAssets = {}
    self.trainLength = 0
    self.locomotiveOffset = 0

    self.splinePositionUpdateListeners = {}
    self.splinePositionListenersActive = false

    self.isSellingStationTrain = true

    self.lastLightStateDirection = SellingStationTrain.DIRECTION_PARKED

    return self
end

function SellingStationTrain:load(nodeId, owningPlaceable, xmlFile, baseKey, customEnvironment, baseDirectory, xmlFilename)
    self.nodeId = nodeId
    self.nodeName = getName(nodeId)

    self.customEnvironment = customEnvironment
    self.baseDirectory = baseDirectory

    if owningPlaceable == nil then
        if owningPlaceable == nil and xmlFilename == nil then
            xmlFilename = self.nodeName
        end

        g_logManager:xmlError(xmlFilename, "Could not load Selling Station Train, parent placeable is missing!")

        return false
    end

    local storeItem = g_storeManager:getItemByXMLFilename(owningPlaceable.configFileName)

    if storeItem ~= nil and (storeItem.canBeSold or storeItem.showInStore) then
        if storeItem.canBeSold then
            g_logManager:xmlError(storeItem.xmlFilename, "Could not load Selling Station Train! Store item %s 'canBeSold' tag must be set 'false' in XML '%s'.", storeItem.name, storeItem.xmlFilename)
        end

        if storeItem.showInStore then
            g_logManager:xmlError(storeItem.xmlFilename, "Could not load Selling Station Train! Store item %s 'showInStore' tag must be set 'false' in XML '%s'.", storeItem.name, storeItem.xmlFilename)
        end

        return false
    end

    xmlFilename = xmlFilename or owningPlaceable.configFileName

    self.owningPlaceable = owningPlaceable
    self.sellingStation = owningPlaceable.sellingStation

    if self.sellingStation ~= nil then
        self.isWoodTrain = false
        self.stationName = self.sellingStation.stationName
        self.acceptedFillTypes = self.sellingStation.acceptedFillTypes
    else
        self.isWoodTrain = true
        self.stationName = owningPlaceable.stationName
        self.acceptedFillTypes = {[FillType.WOODCHIPS] = true}
    end

    self.maxNumFarms = 1

    if g_currentMission.missionDynamicInfo.isMultiplayer then
        self.maxNumFarms = FarmManager.MAX_FARM_ID
    end

    self.debugActive = Utils.getNoNil(getXMLBool(xmlFile, baseKey .. "#debugActive"), false)

    self:resetFarmLevels(nil, true)

    self.spline = I3DUtil.indexToObject(self.nodeId, getXMLString(xmlFile, baseKey .. ".spline#node"))

    if self.spline == nil then
        g_logManager:xmlError(xmlFilename, "No spline node found at '%s.spline'!", baseKey)

        return false
    end

    if not getHasClassId(getGeometry(self.spline), ClassIds.SPLINE) then
        g_logManager:xmlError(xmlFilename, "Given node '%s' is not a spline!", getName(self.spline))

        return false
    end

    self.splineLength = getSplineLength(self.spline)
    self.splineYOffset = Utils.getNoNil(getXMLFloat(xmlFile, baseKey .. ".spline#splineYOffset"), 0)

    self.trainTransform = createTransformGroup("trainTransform")
    link(getRootNode(), self.trainTransform)

    local i = 0
    while true do
        local key = string.format("%s.trainAssets.asset(%d)", baseKey, i)
        if not hasXMLProperty(xmlFile, key) then
            break
        end

        local xmlFilename = getXMLString(xmlFile, key .. "#xmlFilename")

        if xmlFilename ~= nil then
            local assetXmlFile = loadXMLFile("trainAssetXML", self.baseDirectory .. xmlFilename)

            if assetXmlFile ~= nil or assetXmlFile ~= 0 then
                local trainAsset = self:loadTrainAssetFromXML(assetXmlFile, xmlFilename, key)

                if trainAsset ~= nil then
                    self.trainLength = self.trainLength + trainAsset.length
                    table.insert(self.trainAssets, trainAsset)
                end

                delete(assetXmlFile)
            end
        end

        i = i + 1
    end

    if #self.fillPlanes > 0 then
        self:setFillPlanes(false)
    else
        self.fillPlanes = nil
    end

    local interactionTrigger = I3DUtil.indexToObject(self.nodeId, getXMLString(xmlFile, baseKey .. ".interactionTrigger#node"))

    if interactionTrigger == nil then
        g_logManager:xmlWarning(xmlFilename, "Missing interaction trigger at '%s.interactionTrigger'!", baseKey)

        return false
    end

    if bitAND(Player.COLLISIONMASK_TRIGGER, getCollisionMask(interactionTrigger)) == 0 then
        g_logManager:xmlWarning(xmlFilename, "Invalid collision mask for interaction trigger '%s'. Bit 20 needs to be set!", getName(interactionTrigger))

        return false
    end

    self.activatable = SellingStationTrainActivatable:new(self, interactionTrigger)
    self.interactionMarker = I3DUtil.indexToObject(self.nodeId, getXMLString(xmlFile, baseKey .. ".interactionTrigger#marker"))

    if self.interactionMarker ~= nil then
        g_currentMission:addTriggerMarker(self.interactionMarker)
    end

    local leverNode = I3DUtil.indexToObject(self.nodeId, getXMLString(xmlFile, baseKey .. ".interactionTrigger.lever#node"))

    if leverNode ~= nil then
        local maxTime = 0

        local rotCurve = AnimCurve:new(linearInterpolator3)
        local drx, dry, drz = getRotation(leverNode)

        local i = 0
        while true do
            local frameKey = string.format("%s.interactionTrigger.lever.keyFrames.keyFrame(%d)", baseKey, i)
            if not hasXMLProperty(xmlFile, frameKey) then
                break
            end

            local keyTime = getXMLFloat(xmlFile, frameKey .. "#seconds")

            if keyTime ~= nil then
                local rx, ry, rz = StringUtil.getVectorFromString(getXMLString(xmlFile, frameKey .. "#rotation"))
                rx, ry, rz = Utils.getNoNilRad(rx, drx), Utils.getNoNilRad(ry, dry), Utils.getNoNilRad(rz, drz)

                local keyframe = {rx, ry, rz}
                keyframe.time = keyTime * 1000

                rotCurve:addKeyframe(keyframe)

                if keyframe.time > maxTime then
                    maxTime = keyframe.time
                end
            end

            i = i + 1
        end

        if maxTime > 0 then
            self.lever = {
                timer = 0,
                active = false,
                node = leverNode,
                maxTime = maxTime,
                rotCurve = rotCurve
            }

            if self.isClient then
                self.samples.lever = g_soundManager:loadSampleFromXML(xmlFile, baseKey .. ".interactionTrigger.lever.sounds", "moving", self.baseDirectory, self.nodeId, 1, AudioGroup.ENVIRONMENT, nil, nil)
            end
        end
    end

    local trainStatusLights = self:loadLights(self.nodeId, xmlFile, baseKey .. ".statusLights", false)

    if #trainStatusLights > 0 then
        self.trainStatusLights = trainStatusLights
    end

    local signalLights = self:loadLights(self.nodeId, xmlFile, baseKey .. ".signalLights", true)

    if #signalLights > 0 then
        self.signalLights = signalLights
    end

    self.minimumWaitTime = 3000
    self.departWaitTime = math.max(Utils.getNoNil(getXMLFloat(xmlFile, baseKey .. "#departWaitTime"), 3) + 3, 6) * 1000
    self.returnWaitTime = math.max(Utils.getNoNil(getXMLFloat(xmlFile, baseKey .. "#returnWaitTime"), 10), 10) * 1000

    self.maxSpeed = math.max(Utils.getNoNil(getXMLFloat(xmlFile, baseKey .. "#maxSpeed"), 20), 10) / 3600
    self.maxMovingSpeed = self.maxSpeed / self.splineLength
    self.maxAcceleration = Utils.getNoNil(getXMLFloat(xmlFile, baseKey .. "#maxAcceleration"), 10) * 1000

    self.trainLengthSplineTime = self.trainLength / self.splineLength
    self.splineBrakingTime = (self.trainLengthSplineTime / #self.trainAssets) * 6

    self.brakingSpeedStart = self.trainLengthSplineTime + self.splineBrakingTime
    self.brakingSpeedEnd = 1 - self.splineBrakingTime

    self:setPositionListeners(xmlFile, baseKey)

    self.networkTimeInterpolator = InterpolationTime:new(1.2)
    self.networkSplineTimeInterpolator = InterpolatorValue:new(0)

    self.notificationText = g_i18n:getText("sellingStationTrain_trainNotification", self.customEnvironment)
    self.trainDirtyFlag = self:getNextDirtyFlag()

    if #self.lights > 0 then
        self:setLightsState()
    else
        self.lights = nil
    end

    self:updateSplineTime(self.trainLengthSplineTime)

    g_messageCenter:subscribe(MessageType.FARM_DELETED, self.onFarmDeleted, self)

    return true
end

function SellingStationTrain:loadTrainAssetFromXML(xmlFile, xmlFilename, assetKey)
    local trainAsset
    local filename = getXMLString(xmlFile, "trainAsset#filename")

    if filename ~= nil then
        local assetRoot = g_i3DManager:loadSharedI3DFile(filename, self.baseDirectory, false, true, false)

        if assetRoot == nil then
            g_logManager:xmlWarning(xmlFilename, "Failed to load train asset '%s' from XML file!", assetKey)

            return false
        end

        local assetTypeName = Utils.getNoNil(getXMLString(xmlFile, "trainAsset#type"), "wagon")

        if assetTypeName == "wagon" or assetTypeName == "locomotive" then
            trainAsset = {
                wheels = {},
                dollies = {},
                filename = filename,
                assetTypeName = assetTypeName,
                node = getChildAt(assetRoot, 0)
            }

            if assetTypeName == "locomotive" then
                if self.locomotive == nil then
                    self.locomotiveOffset = Utils.getNoNil(getXMLFloat(xmlFile, "trainAsset#offset"), 0) / self.splineLength

                    local characterNode = I3DUtil.indexToObject(trainAsset.node, getXMLString(xmlFile, "trainAsset.locomotive.character#node"))

                    if characterNode ~= nil then
                        setVisibility(characterNode, false)

                        local frontNode = I3DUtil.indexToObject(trainAsset.node, getXMLString(xmlFile, "trainAsset.locomotive.character#frontNode"))
                        local backNode = I3DUtil.indexToObject(trainAsset.node, getXMLString(xmlFile, "trainAsset.locomotive.character#backNode"))

                        if frontNode ~= nil and backNode ~= nil then
                            trainAsset.character = {
                                node = characterNode,
                                frontTranslation = {getTranslation(frontNode)},
                                frontRotation = {getRotation(frontNode)},
                                backTranslation = {getTranslation(backNode)},
                                backRotation = {getRotation(backNode)}
                            }
                        end
                    end

                    if self.isClient then
                        local soundsKey = "trainAsset.locomotive.sounds"

                        trainAsset.samples = {
                            motorStart = g_soundManager:loadSampleFromXML(xmlFile, soundsKey, "motorStart", self.baseDirectory, trainAsset.node, 1, AudioGroup.VEHICLE, nil, self),
                            motorStop = g_soundManager:loadSampleFromXML(xmlFile, soundsKey, "motorStop", self.baseDirectory, trainAsset.node, 1, AudioGroup.VEHICLE, nil, self),
                            gearbox = g_soundManager:loadSampleFromXML(xmlFile, soundsKey, "gearbox", self.baseDirectory, trainAsset.node, 0, AudioGroup.VEHICLE, nil, self),
                            honk = g_soundManager:loadSampleFromXML(xmlFile, "trainAsset.locomotive", "honkSound", self.baseDirectory, trainAsset.node, 10, AudioGroup.VEHICLE, nil, self)
                        }

                        if trainAsset.samples.honk ~= nil then
                            trainAsset.honkLoops = trainAsset.samples.honk.loops
                        end

                        trainAsset.motorSamples = {}

                        local i = 0
                        while true do
                            local sample = g_soundManager:loadSampleFromXML(xmlFile, soundsKey, string.format("motor(%d)", i), self.baseDirectory, trainAsset.node, 0, AudioGroup.VEHICLE, nil, self)
                            if sample == nil then
                                break
                            end

                            table.insert(trainAsset.motorSamples, sample)

                            i = i + 1
                        end

                        trainAsset.animationNodes = g_animationManager:loadAnimations(xmlFile, "trainAsset.locomotive.animationNodes", trainAsset.node, self, nil)
                    end

                    self.locomotive = trainAsset
                    g_currentMission:addNodeObject(trainAsset.node, self)
                end
            else
                trainAsset.fillPlanes = {}

                local i = 0
                while true do
                    local fillPlaneKey = string.format("trainAsset.wagon.fillPlanes.fillPlane(%d)", i)
                    if not hasXMLProperty(xmlFile, fillPlaneKey) then
                        break
                    end

                    local fillPlaneNode = I3DUtil.indexToObject(trainAsset.node, getXMLString(xmlFile, fillPlaneKey .. "#node"))

                    if fillPlaneNode ~= nil then
                        local fillPlane = {
                            node = fillPlaneNode,
                            collisionMask = 0,
                            switchCollisionMask = Utils.getNoNil(getXMLBool(xmlFile, fillPlaneKey .. "#switchCollisionMask"), false)
                        }

                        if fillPlane.switchCollisionMask then
                            fillPlane.collisionMask = getCollisionMask(fillPlaneNode)
                        end

                        table.insert(trainAsset.fillPlanes, fillPlane)
                        table.insert(self.fillPlanes, fillPlane)
                    end

                    i = i + 1
                end

                if #trainAsset.fillPlanes == 0 then
                    trainAsset.fillPlanes = nil
                end
            end

            trainAsset.length = math.max(Utils.getNoNil(getXMLFloat(xmlFile, "trainAsset#length"), 1), 1)

            link(self.trainTransform, trainAsset.node)
            delete(assetRoot)

            local i = 0
            while true do
                local wheelKey = string.format("trainAsset.wheels.wheel(%d)", i)
                if not hasXMLProperty(xmlFile, wheelKey) then
                    break
                end

                local wheelNode = I3DUtil.indexToObject(trainAsset.node, getXMLString(xmlFile, wheelKey .. "#node"))

                if wheelNode ~= nil then
                    local wheelRadius = getXMLFloat(xmlFile, wheelKey .. "#wheelRadius")
                    local direction = MathUtil.sign(Utils.getNoNil(getXMLInt(xmlFile, wheelKey .. "#direction"), 1))

                    if wheelRadius == nil then
                        local _, y, _ = getTranslation(wheelNode)
                        wheelRadius = y
                    end

                    table.insert(trainAsset.wheels, {node = wheelNode, radius = wheelRadius, direction = direction})
                end

                i = i + 1
            end

            if #trainAsset.wheels == 0 then
                g_logManager:xmlWarning(xmlFilename, "Failed to find wheels for train asset '%s'!", assetKey)

                return
            end

            if g_currentMission.surfaceSounds ~= nil then
                for _, surfaceSound in pairs(g_currentMission.surfaceSounds) do
                    if surfaceSound.type == "wheel" and surfaceSound.name == "railroad" and surfaceSound.sample ~= nil then
                        local sample = g_soundManager:cloneSample(surfaceSound.sample, trainAsset.node, self)
                        sample.sampleName = surfaceSound.name
                        table.insert(self.surfaceSounds, sample)

                        break
                    end
                end
            end

            local _, _, frontWheels = getTranslation(trainAsset.wheels[1].node)
            local _, _, rearWheels = getTranslation(trainAsset.wheels[#trainAsset.wheels].node)

            trainAsset.frontWheels = frontWheels
            trainAsset.rearWheels = rearWheels

            trainAsset.wheelToWheelLength = math.abs(trainAsset.frontWheels - trainAsset.rearWheels)

            trainAsset.frontWheelsOffset = math.abs(trainAsset.length * 0.5 - trainAsset.frontWheels)
            trainAsset.rearWheelsOffset = math.abs(trainAsset.rearWheels + trainAsset.length * 0.5)

            setTranslation(trainAsset.node, 0, 0, self.trainLength + trainAsset.length * 0.5)

            if hasXMLProperty(xmlFile, "trainAsset.materials") then
                i = 0
                while true do
                    local materialKey = string.format("trainAsset.materials.material(%d)", i)
                    if not hasXMLProperty(xmlFile, materialKey) then
                        break
                    end

                    local baseMaterialNode = I3DUtil.indexToObject(trainAsset.node, getXMLString(xmlFile, materialKey .. "#node"))
                    local refMaterialNode = I3DUtil.indexToObject(trainAsset.node, getXMLString(xmlFile, materialKey .. "#refNode"))

                    if baseMaterialNode ~= nil and refMaterialNode ~= nil then
                        ConfigurationUtil.replaceMaterialRec(trainAsset, trainAsset.node, getMaterial(baseMaterialNode, 0), getMaterial(refMaterialNode, 0))
                    end

                    i = i + 1
                end
            end

            trainAsset.lights = self:loadLights(trainAsset.node, xmlFile, "trainAsset.lights", false)
        end
    else
        g_logManager:xmlWarning(xmlFilename, "Invalid trainAsset type '%s', use either 'locomotive' or 'wagon'!", assetTypeName)
    end

    return trainAsset
end

function SellingStationTrain:loadLights(nodeId, xmlFile, key, isSignalLight)
    local lights = {}

    local i = 0
    while true do
        local lightKey = string.format("%s.light(%d)", key, i)
        if not hasXMLProperty(xmlFile, lightKey) then
            break
        end

        local light = {
            enableDirections = {
                ["waitingReverse"] = false,
                ["reverse"] = false,
                ["parked"] = false,
                ["forward"] = false,
                ["waitingForward"] = false
            },
            node = I3DUtil.indexToObject(nodeId, getXMLString(xmlFile, lightKey .. "#node")),
            intensity = math.max(Utils.getNoNil(getXMLFloat(xmlFile, lightKey .. "#intensity"), 15), 0.1)
        }

        local realNode = I3DUtil.indexToObject(nodeId, getXMLString(xmlFile, lightKey .. "#realNode"))
        local shaderNode = I3DUtil.indexToObject(nodeId, getXMLString(xmlFile, lightKey .. "#shaderNode"))

        local enableDirections = StringUtil.splitString(" ", Utils.getNoNil(getXMLString(xmlFile, lightKey .. "#enableDirections"), SellingStationTrain.DEFAULT_DIRECTIONS))

        for _, enableDirection in ipairs (enableDirections) do
            light.enableDirections[enableDirection] = true
        end

        -- Adds a position listener to change the light state when train has passed in either direction, useful for swapping signals.
        if isSignalLight then
            local positionNode = I3DUtil.indexToObject(nodeId, getXMLString(xmlFile, lightKey .. ".positionListener#node"))

            if positionNode ~= nil then
                local enableDirection = Utils.getNoNil(getXMLString(xmlFile, lightKey .. ".positionListener#enableDirection"), "parked")
                local disableDirection = Utils.getNoNil(getXMLString(xmlFile, lightKey .. ".positionListener#disableDirection"), "parked")

                enableDirection = SellingStationTrain.NAME_TO_DIRECTION[enableDirection] or 0
                disableDirection = SellingStationTrain.NAME_TO_DIRECTION[disableDirection] or 0

                if enableDirection ~= 0 or disableDirection ~= 0 then
                    local offset = Utils.getNoNil(getXMLInt(xmlFile, lightKey .. ".positionListener#offset"), 5)

                    local lightListener = {
                        node = positionNode,
                        linkedLight = light,
                        lightIntensity = light.intensity,
                        enableDirection = enableDirection,
                        disableDirection = disableDirection,
                        currentDirection = 0,
                        train = self,
                        offset = offset / self.splineLength,
                        validTime = self:getSplineTimeFromNode(positionNode, 1)
                    }

                    lightListener.name = string.format("Light Listener - ID %d", #lights + 1)

                    lightListener.onTrainDelete = function(listener, train)
                        train:removeSplinePositionUpdateListener(listener)
                        listener.train = nil
                    end

                    lightListener.onTravelDirectionChange = function(listener, travelDirection)
                        if travelDirection == SellingStationTrain.DIRECTION_PARKED then
                            listener.currentDirection = SellingStationTrain.DIRECTION_PARKED
                        end
                    end

                    lightListener.onSplinePositionTimeUpdate = function(listener, startTime, endTime, movedDistance, travelDirection)
                        if travelDirection ~= listener.currentDirection then
                            if travelDirection > 0 and listener.validTime < (endTime - listener.offset) or travelDirection < 0 and listener.validTime > (startTime + listener.offset) then
                                if travelDirection == listener.enableDirection then
                                    listener.train:setLightState(listener.linkedLight, listener.lightIntensity or 25)
                                elseif travelDirection == listener.disableDirection then
                                    listener.train:setLightState(listener.linkedLight, 0)
                                end

                                listener.currentDirection = travelDirection
                            end
                        end
                    end

                    self:addSplinePositionUpdateListener(lightListener)
                end
            end
        end

        if shaderNode ~= nil then
            light.shaderNodes = {}

            if getHasShaderParameter(shaderNode, "lightControl") then
                table.insert(light.shaderNodes, shaderNode)
            end

            for i = 0, getNumOfChildren(shaderNode) - 1 do
                local childNode = getChildAt(shaderNode, i)

                if getHasShaderParameter(childNode, "lightControl") then
                    table.insert(light.shaderNodes, childNode)
                end
            end

            if #light.shaderNodes == 0 then
                light.shaderNodes = nil
            end
        end

        if realNode ~= nil and getHasClassId(realNode, ClassIds.LIGHT_SOURCE) then
            light.realNode = realNode
        end

        table.insert(lights, light)
        table.insert(self.lights, light)

        i = i + 1
    end

    return lights
end

function SellingStationTrain:setPositionListeners(xmlFile, key)
    if xmlFile ~= nil and key ~= nil then
        if self.locomotive ~= nil and self.locomotive.samples.honk ~= nil then
            local i = 0
            while true do
                local honkListenerKey = string.format("%s.spline.honkListener(%d)", key, i)
                if not hasXMLProperty(xmlFile, honkListenerKey) then
                    break
                end

                local honkListener = {
                    node = I3DUtil.indexToObject(self.nodeId, getXMLString(xmlFile, honkListenerKey .. "#node"))
                }

                if honkListener.node ~= nil then
                    if self.honkListeners == nil then
                        self.honkListeners = {}
                    end

                    honkListener.forwardDistance = Utils.getNoNil(getXMLInt(xmlFile, honkListenerKey .. "#forwardDistance"), 20)
                    honkListener.reverseDistance = Utils.getNoNil(getXMLInt(xmlFile, honkListenerKey .. "#reverseDistance"), 20)

                    honkListener.delay = math.max(Utils.getNoNil(getXMLFloat(xmlFile, honkListenerKey .. "#delay"), 0), 0) * 1000
                    honkListener.loops = getXMLInt(xmlFile, honkListenerKey .. "#loops")

                    honkListener.train = self
                    honkListener.active = false

                    honkListener.name = string.format("Honk Listener - ID %d", #self.honkListeners + 1)

                    honkListener.validTime = self:getSplineTimeFromNode(honkListener.node)
                    honkListener.startTime = honkListener.validTime - (honkListener.forwardDistance / self.splineLength)
                    honkListener.endTime = honkListener.validTime + (honkListener.reverseDistance / self.splineLength)

                    honkListener.onTrainDelete = function(listener, train)
                        train:removeSplinePositionUpdateListener(listener)
                        listener.train = nil
                    end

                    honkListener.onSplinePositionTimeUpdate = function(listener, startTime, endTime, movedDistance, travelDirection)
                        if listener.train ~= nil then
                            startTime = SplineUtil.getValidSplineTime(startTime)
                            endTime = SplineUtil.getValidSplineTime(endTime)

                            local active = listener.startTime < startTime and endTime < listener.endTime

                            if listener.active ~= active then
                                listener.active = active
                                listener.train:setHonkState(listener.active, listener.delay, listener.loops)
                            end
                        end
                    end

                    self:addSplinePositionUpdateListener(honkListener)

                    table.insert(self.honkListeners, honkListener)
                end

                i = i + 1
            end
        end
    end

    if self.debugActive then
        local debugListener = {
            name = "Debug Position Listener",
            isDebugListener = true,
            train = self
        }

        debugListener.onTrainDelete = function(listener, train)
            train:removeSplinePositionUpdateListener(listener)
            listener.train = nil
        end

        debugListener.onSplinePositionTimeUpdate = function(listener, startTime, endTime, movedDistance, travelDirection)
            g_currentMission:addExtraPrintText(string.format("Front (%s): %.4f", listener.train.stationName, startTime - listener.train.locomotiveOffset))
            g_currentMission:addExtraPrintText(string.format("Rear (%s): %.4f", listener.train.stationName, endTime - listener.train.locomotiveOffset))
            g_currentMission:addExtraPrintText(string.format("Moved Distance (%s): %.4f", listener.train.stationName, movedDistance))
            g_currentMission:addExtraPrintText(string.format("Speed (%s): %.2f", listener.train.stationName, listener.train:getLastSpeed()))
            g_currentMission:addExtraPrintText(string.format("Direction (%s): %d", listener.train.stationName, travelDirection))
        end

        self:addSplinePositionUpdateListener(debugListener)
    end
end

function SellingStationTrain:delete()
    g_messageCenter:unsubscribeAll(self)

    if self.splinePositionListenersActive then
        for _, listener in ipairs(self.splinePositionUpdateListeners) do
            if listener.onTrainDelete ~= nil then
                listener:onTrainDelete(self)
            end
        end
    end

    if self.locomotive ~= nil then
        g_currentMission:removeNodeObject(self.locomotive)
    end

    if self.isClient then
        g_soundManager:deleteSamples(self.samples)

        if self.locomotive ~= nil then
            g_soundManager:deleteSamples(self.locomotive.samples)
            g_soundManager:deleteSamples(self.locomotive.motorSamples)

            g_animationManager:deleteAnimations(self.locomotive.animationNodes)
        end
    end

    if #self.trainAssets > 0 then
        for _, trainAsset in ipairs(self.trainAssets) do
            g_i3DManager:releaseSharedI3DFile(trainAsset.filename, self.baseDirectory, true)
        end

        self.trainAssets = {}
    end

    if self.activatable ~= nil then
        self.activatable:delete()
        self.activatable = nil
    end

    if self.interactionMarker ~= nil then
        g_currentMission:removeTriggerMarker(self.interactionMarker)
        self.interactionMarker = nil
    end

    if self.trainTransform ~= nil then
        delete(self.trainTransform)
    end

    SellingStationTrain:superClass().delete(self)
end

function SellingStationTrain:loadFromXMLFile(xmlFile, key, resetVehicles)
    local travelDirection = MathUtil.clamp(Utils.getNoNil(getXMLFloat(xmlFile, key .. "#travelDirection"), self.travelDirection), -1, 1)

    if travelDirection ~= SellingStationTrain.DIRECTION_PARKED then
        local splineTime = Utils.getNoNil(getXMLFloat(xmlFile, key .. "#splineTime"), self.splineTime)
        self.lastMovingSpeed = Utils.getNoNil(getXMLFloat(xmlFile, key .. "#lastMovingSpeed"), self.lastMovingSpeed)

        self:setTravelDirection(travelDirection, false, nil, false)
        self.waitTime = Utils.getNoNil(getXMLFloat(xmlFile, key .. "#waitTime"), self:getCurrentWaitTime())

        self:updateSplineTime(splineTime)
    end

    local i = 0

    while true do
        local farmKey = string.format("%s.farm(%d)", key, i)
        if not hasXMLProperty(xmlFile, farmKey) then
            break
        end

        local farmId = getXMLInt(xmlFile, farmKey .. "#id")

        if self:getIsValidFarmId(farmId) and g_farmManager:getFarmById(farmId) ~= nil then
            local j = 0

            while true do
                local fillTypeKey = string.format("%s.fillType(%d)", farmKey, j)
                if not hasXMLProperty(xmlFile, fillTypeKey) then
                    break
                end

                local fillTypeStr = getXMLString(xmlFile, fillTypeKey .. "#name")
                local fillTypeIndex = g_fillTypeManager:getFillTypeIndexByName(fillTypeStr)

                if fillTypeIndex ~= nil then
                    if self:getIsFillTypeAllowed(fillTypeIndex) then
                        local fillLevel = math.max(Utils.getNoNil(getXMLFloat(xmlFile, fillTypeKey .. "#fillLevel"), 0), 0)
                        local price = math.max(Utils.getNoNil(getXMLFloat(xmlFile, fillTypeKey .. "#soldPrice"), 0), 0)

                        self:setFillLevel(farmId, fillTypeIndex, fillLevel, 0, price, false)
                    else
                        g_logManager:warning(" [SellingStationTrain]  Filltype '%s' not supported by train at '%s'!", fillTypeStr, self.stationName)
                    end
                else
                    g_logManager:error(" [SellingStationTrain]  Invalid filltype '%s' can not be loaded to train at '%s'!", fillTypeStr, self.stationName)
                end

                j = j + 1
            end
        end

        i = i + 1
    end

    return true
end

function SellingStationTrain:saveToXMLFile(xmlFile, key, usedModNames)
    setXMLInt(xmlFile, key .. "#travelDirection", self.travelDirection)

    if self.travelDirection ~= SellingStationTrain.DIRECTION_PARKED then
        setXMLFloat(xmlFile, key .. "#splineTime", self.splineTime)
        setXMLFloat(xmlFile, key .. "#lastMovingSpeed", self.lastMovingSpeed)

        setXMLFloat(xmlFile, key .. "#waitTime", self.waitTime or 0)

        if self.startingUserName ~= nil and self.startingUserName ~= "" then
            setXMLString(xmlFile, key .. "#startingUserName", self.startingUserName)
        end
    end

    local index = 0

    for farmId, fillTypes in pairs (self.fillLevels) do
        local farm = g_farmManager:getFarmById(farmId)

        if farm ~= nil then
            local fillTypeCount = 0
            local farmKey = string.format("%s.farm(%d)", key, index)

            for fillTypeIndex, fillLevel in pairs (fillTypes) do
                if fillLevel > 0 then
                    local fillTypeKey = string.format("%s.fillType(%d)", farmKey, fillTypeCount)

                    setXMLString(xmlFile, fillTypeKey .. "#name", g_fillTypeManager:getFillTypeNameByIndex(fillTypeIndex))
                    setXMLFloat(xmlFile, fillTypeKey .. "#fillLevel", fillLevel)

                    if self.isWoodTrain then
                        setXMLFloat(xmlFile, fillTypeKey .. "#soldPrice", self:getSoldPrice(farmId, fillTypeIndex))
                    end

                    fillTypeCount = fillTypeCount + 1
                end
            end

            if fillTypeCount > 0 then
                setXMLInt(xmlFile, farmKey .. "#id", farmId)
                setXMLString(xmlFile, farmKey .. "#name", farm.name)

                index = index + 1
            end
        end
    end
end

function SellingStationTrain:readStream(streamId, connection)
    SellingStationTrain:superClass().readStream(self, streamId, connection)

    if connection:getIsServer() then
        local travelDirection = streamReadUIntN(streamId, 2) - 1
        local splineTime = streamReadFloat32(streamId)

        local userId = streamReadInt8(streamId)
        self:setTravelDirection(travelDirection, false, userId, false)

        self.networkSplineTimeInterpolator:setValue(splineTime)
        self.networkTimeInterpolator:reset()

        self:updateSplineTime(splineTime)

        for farmId = 1, self.maxNumFarms do
            local numFillTypes = streamReadUInt8(streamId)

            if numFillTypes > 0 then
                for i = 1, numFillTypes do
                    local fillLevel
                    local fillTypeIndex = streamReadUIntN(streamId, FillTypeManager.SEND_NUM_BITS)

                    if streamReadBool(streamId) then
                        fillLevel = streamReadFloat32(streamId)
                    end

                    self:setFillLevel(farmId, fillTypeIndex, fillLevel, nil, nil, false)
                end
            end
        end

        self:setNeedsUpdate()
    end
end

function SellingStationTrain:writeStream(streamId, connection)
    SellingStationTrain:superClass().writeStream(self, streamId, connection)

    if not connection:getIsServer() then
        local travelDirection = MathUtil.sign(self.travelDirection)

        streamWriteUIntN(streamId, travelDirection + 1, 2)
        streamWriteFloat32(streamId, self.splineTimeSent)

        local userId = self.startingUserId or 0
        streamWriteInt8(streamId, userId)

        for farmId = 1, self.maxNumFarms do
            local numFillTypes = SellingStationTrain.getTableSize(self.fillLevels[farmId])

            streamWriteUInt8(streamId, numFillTypes)

            if numFillTypes > 0 then
                for fillTypeIndex, fillLevel in pairs (self.fillLevels[farmId]) do
                    streamWriteUIntN(streamId, fillTypeIndex, FillTypeManager.SEND_NUM_BITS)

                    if streamWriteBool(streamId, fillLevel > 0) then
                        streamWriteFloat32(streamId, fillLevel)
                    end
                end
            end
        end
    end
end

function SellingStationTrain:readUpdateStream(streamId, timestamp, connection)
    SellingStationTrain:superClass().readUpdateStream(self, streamId, timestamp, connection)

    if connection:getIsServer() then
        if streamReadBool(streamId) then
            local splineTime = streamReadFloat32(streamId)

            self.networkTimeInterpolator:startNewPhaseNetwork()
            self.networkSplineTimeInterpolator:setTargetValue(splineTime)

            self:setNeedsUpdate()
        end
    end
end

function SellingStationTrain:writeUpdateStream(streamId, connection, dirtyMask)
    SellingStationTrain:superClass().writeUpdateStream(self, streamId, connection, dirtyMask)

    if not connection:getIsServer() then
        if streamWriteBool(streamId, bitAND(dirtyMask, self.trainDirtyFlag) ~= 0) then
            streamWriteFloat32(streamId, self.splineTimeSent)
        end
    end
end

function SellingStationTrain:update(dt)
    local movedDistance = 0

    if self.isServer then
        if self.travelDirection ~= 0 then
            if self.waitTime > self:getCurrentWaitTime() then
                if self.brakingSpeed > 0 then
                    self.lastMovingSpeed = self.startBrakeSpeed * self.brakingSpeed
                else
                    self.lastMovingSpeed = MathUtil.clamp(self.lastMovingSpeed - dt / self.maxAcceleration, -1, 1)
                end

                movedDistance = self:updateSplineTime(self.splineTime + self.maxMovingSpeed * -self.lastMovingSpeed * self.travelDirection * dt)
            else
                self.waitTime = self.waitTime + dt
            end

            self.needsUpdate = true
        end

        self.lastSpeed = (movedDistance / dt) * (self.travelDirection <= 0 and 1 or -1)

        if self.timeSinceFillTypesAdded > 0 then
            self.timeSinceFillTypesAdded = self.timeSinceFillTypesAdded - 1

            if self.timeSinceFillTypesAdded == 0 then
                local currentFarmId = g_currentMission:getFarmId()

                for farmId = 1, self.maxNumFarms do
                    if g_dedicatedServerInfo == nil and farmId == currentFarmId then
                        for fillTypeIndex, amountAdded in pairs (self.lastAdded[farmId]) do
                            self:showFillTypeAddedNotification(fillTypeIndex, amountAdded)
                        end
                    end

                    self.lastAdded[farmId] = {}
                end

                g_server:broadcastEvent(SellingStationTrainFillLevelEvent:new(self), nil, nil, self)
            else
                self.needsUpdate = true
            end
        end
    else
        self.networkTimeInterpolator:update(dt)
        movedDistance = self:updateSplineTime(self.networkSplineTimeInterpolator:getInterpolatedValue(self.networkTimeInterpolator:getAlpha()))

        self.lastSpeed = (movedDistance / dt) * (self.travelDirection <= 0 and 1 or -1)

        if self.networkTimeInterpolator:isInterpolating() then
            self.needsUpdate = true
        end
    end

    if self.locomotiveStartTime > 0 and self.locomotiveStartTime < g_currentMission.time then
        self.locomotiveStartTime = -1

        self:setLocomotiveCharacter(true)
        self:setLocomotiveState(true)

        self:setLightsState(SellingStationTrain.DIRECTION_FOWARD_WAITING)
        self.lastLightStateDirection = SellingStationTrain.DIRECTION_FOWARD_WAITING
    end

    if self.lastSpeed > 0 and self.lastLightStateDirection ~= self.travelDirection then
        self.lastLightStateDirection = self.travelDirection
        self:setLightsState(self.travelDirection)
    end

    if self.lever ~= nil and self.lever.active then
        local lever = self.lever
        local rx, ry, rz = self.lever.rotCurve:get(lever.timer)

        setRotation(lever.node, rx, ry, rz)
        lever.timer = lever.timer + dt

        if lever.timer > lever.maxTime then
            self:setLeverState(false)
        end

        self.needsUpdate = true
    end

    if self.debugActive then
        self:debugUpdate(dt)
    end

    if self.needsUpdate then
        self:raiseActive()
    end
end

function SellingStationTrain:debugUpdate(dt)
    if self.splinePositionListenersActive then
        local usedNodes, textSize = {}, getCorrectTextSize(0.4)

        for _, listener in ipairs (self.splinePositionUpdateListeners) do
            if listener.node ~= nil then
                if listener.startTime ~= nil and listener.endTime ~= nil then
                    usedNodes[listener.node] = (usedNodes[listener.node] or -textSize) + textSize

                    local x, y, z = getSplinePosition(listener.train.spline, listener.startTime)
                    local dirX, dirY, dirZ = getSplineDirection(listener.train.spline, listener.startTime)
                    local upX, upY, upZ = localDirectionToWorld(listener.node, 0, 1, 0)

                    DebugUtil.drawDebugGizmoAtWorldPos(x, y + usedNodes[listener.node], z, dirX, dirY, dirZ, upX, upY, upZ, listener.name .. "  (Start)", false)

                    x, y, z = getSplinePosition(listener.train.spline, listener.endTime)
                    dirX, dirY, dirZ = getSplineDirection(listener.train.spline, listener.endTime)

                    DebugUtil.drawDebugGizmoAtWorldPos(x, y + usedNodes[listener.node], z, dirX, dirY, dirZ, upX, upY, upZ, listener.name .. "  (End)", false)
                elseif listener.validTime ~= nil then
                    usedNodes[listener.node] = (usedNodes[listener.node] or -textSize) + textSize

                    local x, y, z = getSplinePosition(listener.train.spline, listener.validTime)
                    local dirX, dirY, dirZ = getSplineDirection(listener.train.spline, listener.validTime)
                    local upX, upY, upZ = localDirectionToWorld(listener.node, 0, 1, 0)

                    DebugUtil.drawDebugGizmoAtWorldPos(x, y + usedNodes[listener.node], z, dirX, dirY, dirZ, upX, upY, upZ, listener.name, false)
                end
            end
        end
    end

    local startTime = self.splineTime + self.locomotiveOffset
    local endTime = startTime - self.trainLengthSplineTime

    local x, y, z = getSplinePosition(self.spline, startTime)
    local dirX, dirY, dirZ = getSplineDirection(self.spline, startTime)
    local upX, upY, upZ = localDirectionToWorld(self.locomotive.node, 0, 1, 0)

    DebugUtil.drawDebugGizmoAtWorldPos(x, y, z, dirX, dirY, dirZ, upX, upY, upZ, "Front", false)

    x, y, z = getSplinePosition(self.spline, endTime)
    dirX, dirY, dirZ = getSplineDirection(self.spline, endTime)

    DebugUtil.drawDebugGizmoAtWorldPos(x, y, z, dirX, dirY, dirZ, upX, upY, upZ, "Rear", false)

    self.needsUpdate = true
end

function SellingStationTrain:updateSplineTime(splineTime)
    splineTime = MathUtil.clamp(splineTime, self.trainLengthSplineTime, 1)

    if splineTime ~= self.splineTime then
        local offset, frontWheel, currentTime = nil, nil, splineTime
        local distance = (self.splineTime - splineTime) * self.splineLength
        local startPosX, startPosY, startPosZ = getSplinePosition(self.spline, splineTime)

        for _, asset in ipairs(self.trainAssets) do
            if offset ~= nil then
                startPosX, startPosY, startPosZ, currentTime = self:getSplinePositionAndTime(currentTime, asset.frontWheelsOffset + offset)
            end

            local endPosX, endPosY, endPosZ, endTime = self:getSplinePositionAndTime(currentTime, asset.wheelToWheelLength, 0)

            currentTime = endTime
            offset = asset.rearWheelsOffset
            frontWheel = asset.frontWheels

            local dirX, dirY, dirZ = endPosX - startPosX, endPosY - startPosY, endPosZ - startPosZ
            setDirection(asset.node, dirX, dirY, dirZ, 0, 1, 0)

            dirX, dirY, dirZ = MathUtil.vector3Normalize(dirX, dirY, dirZ)

            startPosX, startPosY, startPosXZ = startPosX + dirX * frontWheel, startPosY + dirY * frontWheel, startPosZ + dirZ * frontWheel
            setTranslation(asset.node, startPosX, startPosY + self.splineYOffset, startPosXZ)

            for _, wheel in ipairs(asset.wheels) do
                rotate(wheel.node, (1 / wheel.radius) * distance * wheel.direction, 0, 0)
            end
        end

        if self.splinePositionListenersActive then
            local offsetStartTime = splineTime + self.locomotiveOffset

            for _, listener in pairs (self.splinePositionUpdateListeners) do
                listener:onSplinePositionTimeUpdate(offsetStartTime, offsetStartTime - self.trainLengthSplineTime, distance, self.travelDirection)
            end
        end

        self.splineTime = splineTime

        if self.isServer then
            local canRaiseDirtyFlags = true

            if self.travelDirection == SellingStationTrain.DIRECTION_FOWARD then
                if self.splineTime >= self.brakingSpeedEnd then
                    self:setBrakingSpeed((1 - self.splineTime) / self.splineBrakingTime)
                end

                if self.splineTime == 1 then
                    self:setTravelDirection(SellingStationTrain.DIRECTION_REVERSE, false, nil, true)
                    canRaiseDirtyFlags = false
                end
            elseif self.travelDirection == SellingStationTrain.DIRECTION_REVERSE then
                if self.splineTime <= self.brakingSpeedStart then
                    self:setBrakingSpeed((self.splineTime - self.trainLengthSplineTime) / self.splineBrakingTime)
                end

                if self.splineTime == self.trainLengthSplineTime then
                    self:setTravelDirection(SellingStationTrain.DIRECTION_PARKED, false, nil, true)
                    canRaiseDirtyFlags = false
                end
            end

            if canRaiseDirtyFlags and self.splineTimeSent ~= splineTime then
                self.splineTimeSent = self.splineTime
                self:raiseDirtyFlags(self.trainDirtyFlag)
            end
        end

        return distance
    end

    return 0
end

function SellingStationTrain:setTravelDirection(direction, isStarting, userId, sendEvent)
    direction = MathUtil.sign(direction)

    if direction ~= self.travelDirection then
        self.travelDirection = direction

        if g_server ~= nil and sendEvent then
            g_server:broadcastEvent(SellingStationTrainSetDirectionEvent:new(self, direction), nil, nil, self)
        end

        self.waitTime = 0
        self.lastSpeed = 0
        self.brakingSpeed = 0
        self.startBrakeSpeed = 0

        if self.splinePositionListenersActive then
            for _, listener in pairs (self.splinePositionUpdateListeners) do
                if listener.onTravelDirectionChange ~= nil then
                    listener:onTravelDirectionChange(direction)
                end
            end
        end

        if direction == SellingStationTrain.DIRECTION_REVERSE then
            if self.isServer and self.splineTime == 1 then
                for farmId, fillTypes in pairs (self.fillLevels) do
                    if g_farmManager:getFarmById(farmId) ~= nil then

                        for fillTypeIndex, fillLevel in pairs (fillTypes) do
                            if fillLevel > 0 then
                                if self.sellingStation ~= nil then
                                    self.sellingStation:sellFillType(farmId, fillLevel, fillTypeIndex, ToolType.DISCHARGEABLE)
                                else
                                    local soldPrice = self:getSoldPrice(farmId, fillTypeIndex)

                                    if soldPrice > 0 then
                                        g_currentMission:farmStats():updateStats("woodTonsSold", fillLevel) -- Same as previous game versions?
                                        g_currentMission:addMoney(soldPrice, farmId, MoneyType.SOLD_WOOD, true, true)
                                    end
                                end
                            end
                        end

                    end
                end
            end

            self:setFillPlanes(false)

            self:resetFarmLevels(nil, true)
            self:setLocomotiveCharacter(true)

            self:setLightsState(SellingStationTrain.DIRECTION_REVERSE_WAITING)
            self.lastLightStateDirection = SellingStationTrain.DIRECTION_REVERSE_WAITING
        elseif direction == SellingStationTrain.DIRECTION_FOWARD then
            self:setNotAllowedWarningText(g_i18n:getText("sellingStationTrain_noTrainAvailable", self.customEnvironment), true)
            self:setStartingUserNameById(userId)

            self:setFillPlanes(true)

            if isStarting then
                self.locomotiveStartTime = g_currentMission.time + self.minimumWaitTime

                self:updateSplineTime(self.trainLengthSplineTime)
                self:setLeverState(true)

                if self.maxNumFarms > 1 or self.debugActive then
                    if self.startingUserName ~= nil and self.startingUserName ~= "" then
                        g_currentMission:addIngameNotification(FSBaseMission.INGAME_NOTIFICATION_OK, string.format(g_i18n:getText("sellingStationTrain_trainStartedBy", self.customEnvironment), self.startingUserName, self.stationName))
                    end
                end
            else
                self:setLocomotiveCharacter(true)
                self:setLocomotiveState(true)

                self:setLightsState(SellingStationTrain.DIRECTION_FOWARD)
                self.lastLightStateDirection = SellingStationTrain.DIRECTION_FOWARD

                self.locomotiveStartTime = -1
            end
        elseif direction == SellingStationTrain.DIRECTION_PARKED then
            self:setLocomotiveState(false)
            self:setLocomotiveCharacter(false)

            self:setNotAllowedWarningText(nil, true)
            self:setStartingUserNameById(nil)

            self:setLightsState(SellingStationTrain.DIRECTION_PARKED)
            self.lastLightStateDirection = SellingStationTrain.DIRECTION_PARKED

            self.locomotiveStartTime = -1
        end

        self:setNeedsUpdate()

        return true
    end

    return false
end

function SellingStationTrain:setBrakingSpeed(brakingSpeed)
    if self.startBrakeSpeed == 0 then
        self.startBrakeSpeed = self.lastMovingSpeed
    end

    self.brakingSpeed = math.max(0.05, brakingSpeed)
end

function SellingStationTrain:setLocomotiveState(active)
    if self.locomotive ~= nil and self.locomotive.active ~= active then
        self.locomotive.active = Utils.getNoNil(active, false)

        if self.isClient then
            g_soundManager:stopSample(self.locomotive.samples.honk)

            if self.locomotive.active then
                g_soundManager:stopSample(self.locomotive.samples.motorStop)
                g_soundManager:playSample(self.locomotive.samples.motorStart)
                g_soundManager:playSamples(self.locomotive.motorSamples, 0, self.locomotive.samples.motorStart)
                g_soundManager:playSample(self.locomotive.samples.gearbox, 0, self.locomotive.samples.motorStart)
                g_soundManager:playSamples(self.surfaceSounds, 0, self.locomotive.samples.motorStart)

                g_animationManager:startAnimations(self.locomotive.animationNodes)
            else
                g_soundManager:stopSample(self.locomotive.samples.motorStart)
                g_soundManager:playSample(self.locomotive.samples.motorStop)
                g_soundManager:stopSamples(self.locomotive.motorSamples)
                g_soundManager:stopSample(self.locomotive.samples.gearbox)
                g_soundManager:stopSamples(self.surfaceSounds)

                g_animationManager:stopAnimations(self.locomotive.animationNodes)
            end
        end
    end
end

function SellingStationTrain:setLocomotiveCharacter(active)
    if self.locomotive ~= nil and self.locomotive.character ~= nil then
        local character = self.locomotive.character

        active = Utils.getNoNil(active, false)

        if active then
            if self.travelDirection == SellingStationTrain.DIRECTION_REVERSE then
                setTranslation(character.node, unpack(character.backTranslation))
                setRotation(character.node, unpack(character.backRotation))
            else
                setTranslation(character.node, unpack(character.frontTranslation))
                setRotation(character.node, unpack(character.frontRotation))
            end
        end

        setVisibility(character.node, active)
    end
end

function SellingStationTrain:setLightsState(direction, lights)
    lights = lights or self.lights

    if lights ~= nil then
        local directionName = self:getDirectionName(direction)

        for _, light in ipairs (lights) do
            self:setLightState(light, light.enableDirections[directionName] and light.intensity or 0)
        end
    end
end

function SellingStationTrain:setLightState(light, intensity)
    if light ~= nil then
        if light.node ~= nil then
            setVisibility(light.node, intensity > 0)
        end

        if light.shaderNodes ~= nil then
            for _, shaderNode in ipairs (light.shaderNodes) do
                local _, y, z, w = getShaderParameter(shaderNode, "lightControl")
                setShaderParameter(shaderNode, "lightControl", intensity, y, z, w, false)
            end
        end

        if light.realNode ~= nil then
            setVisibility(light.realNode, intensity > 0)
        end
    end
end

function SellingStationTrain:setNotAllowedWarningText(text, force)
    if self.notAllowedWarningText ~= text or force then
        self.notAllowedWarningText = text

        if self.sellingStation ~= nil then
            for _, unloadTrigger in pairs(self.sellingStation.unloadTriggers) do
                unloadTrigger.notAllowedWarningText = text
            end
        end
    end
end

function SellingStationTrain:setLeverState(active)
    if self.lever ~= nil then
        self.lever.active = Utils.getNoNil(active, false)
        self.lever.timer = 0

        if self.isClient then
            if self.lever.active then
                if not self.lever.samplePlaying then
                    g_soundManager:playSample(self.samples.lever)
                    self.lever.samplePlaying = true
                end
            else
                if self.lever.samplePlaying then
                    g_soundManager:stopSample(self.samples.lever)
                    self.lever.samplePlaying = false
                end
            end
        end
    end
end

function SellingStationTrain:setHonkState(active, delay, loops)
    if self.isClient then
        if self.locomotive ~= nil and self.locomotive.samples ~= nil then
            if active then
                if not g_soundManager:getIsSamplePlaying(self.locomotive.samples.honk) then

                    if self.locomotive.honkLoops ~= nil then
                        self.locomotive.samples.honk.loops = Utils.getNoNil(loops, self.locomotive.honkLoops)
                    end

                    g_soundManager:playSample(self.locomotive.samples.honk, delay)

                    return true
                end
            else
                g_soundManager:stopSample(self.locomotive.samples.honk)

                return true
            end
        end
    end

    return false
end

function SellingStationTrain:setNeedsUpdate()
    if not self.needsUpdate then
        self.needsUpdate = true
        self:raiseActive()
    end
end

function SellingStationTrain:setStartingUserNameById(userId)
    local nickname = ""

    if userId ~= nil and userId > 0 then
        local user = g_currentMission.userManager:getUserByUserId(userId)

        if user ~= nil then
            nickname = user:getNickname()
        end
    end

    self.startingUserName = nickname
    self.startingUserId = userId or 0
end

function SellingStationTrain:showFillTypeAddedNotification(fillTypeIndex, amountAdded)
    if amountAdded > 0 then
        if self.isWoodTrain then
            g_currentMission:addIngameNotification(FSBaseMission.INGAME_NOTIFICATION_OK, string.format("+ %s (%s - %s)", self:getFormatedWeightFromVolume(amountAdded), self.stationName, self.notificationText))
        else
            local title = " "
            local fillType = g_fillTypeManager:getFillTypeByIndex(fillTypeIndex)

            if fillType ~= nil then
                title = string.format(" (%s) ", fillType.title)
            end

            g_currentMission:addIngameNotification(FSBaseMission.INGAME_NOTIFICATION_OK, string.format("+ %d %s%s(%s - %s)", math.floor(amountAdded + 0.5), g_i18n:getVolumeUnit(true), title, self.stationName, self.notificationText))
        end
    end
end

function SellingStationTrain:addFillLevelToTrain(farmId, fillTypeIndex, fillLevel, soldPrice, pieces)
    if fillLevel > 0 and self:getIsValidFarmId(farmId) then
        local currentFillLevel = self:getFillLevel(farmId, fillTypeIndex)
        local currentLastAdded = self:getLastAddedFillLevel(farmId, fillTypeIndex)

        if self.isWoodTrain and soldPrice > 0 then
            local currentSoldPrice = self:getSoldPrice(farmId, fillTypeIndex)
            soldPrice = currentSoldPrice + soldPrice
        end

        self:setFillLevel(farmId, fillTypeIndex, currentFillLevel + fillLevel, currentLastAdded + fillLevel, soldPrice, false)

        self.timeSinceFillTypesAdded = 30
        self:setNeedsUpdate()

        return 0
    end

    return fillLevel
end

function SellingStationTrain:setFillLevel(farmId, fillTypeIndex, fillLevel, lastAdded, price, showFillTypeChange)
    if fillTypeIndex ~= nil then
        self.fillLevels[farmId][fillTypeIndex] = fillLevel

        if self.isServer then
            self.lastAdded[farmId][fillTypeIndex] = lastAdded

            if self.isWoodTrain then
                self.soldPrices[farmId][fillTypeIndex] = price
            end
        else
            if showFillTypeChange and fillLevel ~= nil then
                self:showFillTypeAddedNotification(fillTypeIndex, lastAdded or 0)
            end
        end
    end
end

function SellingStationTrain:setFillPlanes(active)
    if self.fillPlanes ~= nil then
        active = active or (self.travelDirection > SellingStationTrain.DIRECTION_PARKED)

        for _, fillPlane in pairs (self.fillPlanes) do
            setVisibility(fillPlane.node, active)

            if fillPlane.switchCollisionMask then
                setCollisionMask(fillPlane.node, active and fillPlane.collisionMask or 0)
            end
        end
    end
end

function SellingStationTrain:getFillLevel(farmId, fillTypeIndex)
    return self.fillLevels[farmId][fillTypeIndex] or 0
end

function SellingStationTrain:getLastAddedFillLevel(farmId, fillTypeIndex)
    return self.lastAdded[farmId][fillTypeIndex] or 0
end

function SellingStationTrain:getSoldPrice(farmId, fillTypeIndex)
    return self.soldPrices[farmId][fillTypeIndex] or 0
end

function SellingStationTrain:resetFarmLevels(farmId, resetAll)
    if resetAll then
        for i = 1, self.maxNumFarms do
            self.fillLevels[i] = {}
            self.lastAdded[i] = {}
            self.soldPrices[i] = {}
        end

        return true
    end

    if self:getIsValidFarmId(farmId) then
        self.fillLevels[farmId] = {}
        self.lastAdded[farmId] = {}
        self.soldPrices[farmId] = {}

        return true
    end

    return false
end

function SellingStationTrain:getIsValidFarmId(farmId)
    return farmId ~= nil and farmId > FarmManager.SPECTATOR_FARM_ID and farmId <= self.maxNumFarms
end

function SellingStationTrain:getIsFillTypeAllowed(fillTypeIndex)
    if self.sellingStation ~= nil then
        return self.sellingStation:getIsFillTypeAllowed(fillTypeIndex)
    end

    return fillTypeIndex == FillType.WOODCHIPS
end

function SellingStationTrain:getAcceptedFillTypes()
    if self.sellingStation ~= nil then
        return self.sellingStation.acceptedFillTypes
    end

    return self.acceptedFillTypes
end

function SellingStationTrain:addSplinePositionUpdateListener(listener)
    local added, listLength = ListUtil.addElementToList(self.splinePositionUpdateListeners, listener)
    self.splinePositionListenersActive = listLength > 0

    return added, listLength
end

function SellingStationTrain:removeSplinePositionUpdateListener(listener)
    if ListUtil.removeElementFromList(self.splinePositionUpdateListeners, listener) then
        self.splinePositionListenersActive = #self.splinePositionUpdateListeners > 0

        return true
    end

    return false
end

function SellingStationTrain:onFarmDeleted(farmId)
    self:resetFarmLevels(farmId, false)
end

function SellingStationTrain:getSplineTimeFromNode(node)
    local lastDistance, validTime = math.huge, 0

    for t = 0, 1, 0.5 / self.splineLength do
        local x, y, z = getWorldTranslation(node)
        local sx, sy, sz = getSplinePosition(self.spline, t)
        local distance = MathUtil.vector3Length(sx - x, sy - y, sz - z)

        if distance < lastDistance then
            lastDistance = distance
            validTime = t
        end
    end

    return SplineUtil.getValidSplineTime(validTime)
end

function SellingStationTrain:getFormatedWeightFromVolume(weight, prefix)
    prefix = prefix or ""
    weight = weight * 1000 * 0.47 -- Use 0.47 to get close to the 'getMass()' value.

    if g_i18n.useMiles then
        return string.format("%s%1.1f lbs", prefix, weight * 2.20462)
    end

    return string.format("%s%1.0f kg", prefix, weight)
end

function SellingStationTrain:getLastSpeed()
    return self.lastSpeed * 3600
end

function SellingStationTrain:getDirectionName(direction)
    direction = (direction or self.travelDirection)
    return SellingStationTrain.DIRECTION_P3_TO_NAME[direction + 3] or "parked"
end

function SellingStationTrain:getCurrentWaitTime()
    if self.travelDirection == SellingStationTrain.DIRECTION_REVERSE then
        return self.returnWaitTime
    end

    return self.departWaitTime
end

function SellingStationTrain:getSplinePositionAndTime(currentTime, distance)
    local x, y, z, t = getSplinePositionWithDistance(self.spline, currentTime, distance, false, 0.01)
    return x, y, z, SplineUtil.getValidSplineTime(t)
end

function SellingStationTrain:getIsTrainReadyForDelivery(sellingStation, fillType, farmId, isWoodSellingStation)
    return self:getIsTrainReadyForTravel()
end

function SellingStationTrain:getIsTrainReadyForTravel()
    return self.travelDirection == SellingStationTrain.DIRECTION_PARKED
end

function SellingStationTrain:getTotalFillLevelByFarmId(farmId)
    local totalFillLevel = 0

    if self.fillLevels[farmId] ~= nil then
        for fillTypeIndex, fillLevel in pairs (self.fillLevels[farmId]) do
            totalFillLevel = totalFillLevel + fillLevel
        end
    end

    return totalFillLevel
end

function SellingStationTrain:getCanFarmStartTrain(farmId)
    local canStartTrain, message, dialogType = false, ""
    local farmFillLevels = self.fillLevels[farmId]

    if farmFillLevels ~= nil then
        if self:getIsTrainReadyForTravel() then
            local concatedTitles, volumeUnit = "", g_i18n:getVolumeUnit(true)

            for fillTypeIndex, _ in pairs (self:getAcceptedFillTypes()) do
                local fillType = g_fillTypeManager:getFillTypeByIndex(fillTypeIndex)

                if fillType ~= nil then
                    local fillLevel = self:getFillLevel(farmId, fillTypeIndex)

                    if fillLevel > 0 then
                        canStartTrain = true

                        if self.isWoodTrain then
                            message = "\n" .. self:getFormatedWeightFromVolume(fillLevel, g_i18n:getText("sellingStationTrain_woodPrefix", self.customEnvironment))
                            break
                        else
                            message = string.format("%s\n%s:  %d %s", message, fillType.title, math.floor(fillLevel + 0.5), volumeUnit)
                        end
                    else
                        if self.isWoodTrain then
                            message = g_i18n:getText("sellingStationTrain_woodTrainEmpty", self.customEnvironment)
                            break
                        end
                    end

                    if not canStartTrain then
                        if concatedTitles == "" then
                            concatedTitles = fillType.title
                        else
                            concatedTitles = concatedTitles .. ", " .. fillType.title
                        end
                    end
                end
            end

            if message == "" then
                if concatedTitles ~= "" then
                    message = string.format("%s\n\n%s.", g_i18n:getText("sellingStationTrain_grainTrainEmpty", self.customEnvironment), tostring(concatedTitles))
                else
                    message = g_i18n:getText("warning_actionNotAllowedNow")
                end
            end
        else
            local fillLevelsText = ""

            dialogType = DialogElement.TYPE_LOADING

            for fillTypeIndex, fillLevel in pairs (farmFillLevels) do
                if self.isWoodTrain then
                    fillLevelsText = self:getFormatedWeightFromVolume(fillLevel, g_i18n:getText("sellingStationTrain_woodPrefix", self.customEnvironment))
                    break
                else
                    fillLevelsText = string.format("%s%s:  %d %s\n", fillLevelsText, g_fillTypeManager:getFillTypeNameByIndex(fillTypeIndex), math.floor(fillLevel + 0.5), g_i18n:getVolumeUnit(true))
                end
            end

            if fillLevelsText == "" then
                message = g_i18n:getText("sellingStationTrain_noTrainAvailable", self.customEnvironment)
            else
                message = string.format("%s:\n\n%s", string.upper(g_i18n:getText("fieldJob_jobType_transporting")), fillLevelsText)
            end
        end
    else
        message = g_i18n:getText("warning_actionNotAllowedNow")
    end

    return canStartTrain, message, dialogType
end

function SellingStationTrain:getValidFarmTitle(farmId)
    if self:getIsValidFarmId(farmId) then
        local farm = g_farmManager:getFarmById(farmId)

        if farm ~= nil then
            local farmName = farm.name or "N/A"

            if self.maxNumFarms == 1 then
                if farmName == g_i18n:getText("ui_defaultFarmName") then
                    local user = g_currentMission.userManager:getUserByUserId(g_currentMission.playerUserId)

                    if user ~= nil then
                        farmName = user:getNickname()
                    end
                end
            end

            return string.format("%s:  %s", g_i18n:getText("ui_farm"), farmName)
        end
    end

    return g_i18n:getText("ui_noFarm")
end

function SellingStationTrain:getFarmHasFillLevels(farmId)
    local farmName = self:getValidFarmTitle(farmId)

    if self.fillLevels[farmId] ~= nil then
        for _, fillLevel in pairs (self.fillLevels[farmId]) do
            if fillLevel > 0 then
                return true, farmName
            end
        end
    end

    return false, farmName
end

function SellingStationTrain:onConfirmStartRequest(confirm)
    if confirm then
        local title, text = "", ""

        if self.maxNumFarms > 1 then
            for farmId = 1, self.maxNumFarms do
                if farmId ~= g_currentMission.player.farmId then
                    local hasFillLevels, name = self:getFarmHasFillLevels(farmId)

                    if hasFillLevels then
                        text = string.format("%s\n%s", text, name)
                    end
                end
            end

            if text ~= "" then
                text = g_i18n:getText("sellingStationTrain_confirmStart", self.customEnvironment) .. "\n" .. text
            else
                text = g_i18n:getText("action_sendOffTrain")
                title = g_i18n:getText("button_confirm") .. ":"
            end
        else
            text = g_i18n:getText("action_sendOffTrain")
            title = g_i18n:getText("button_confirm") .. ":"
        end

        g_gui:showYesNoDialog({
            title = title,
            text = text,
            callback = self.onFinaliseStartRequest,
            target = self
        })
    end
end

function SellingStationTrain:onFinaliseStartRequest(confirmStart)
    if confirmStart then
        SellingStationTrainStartRequestEvent.sendEvent(self, g_currentMission.playerUserId)
    end
end

function SellingStationTrain:requestToStartTrain(farmId)
    if self:getIsValidFarmId(farmId) then
        local canStartTrain, message, dialogType = self:getCanFarmStartTrain(farmId)

        if canStartTrain then
            local farmName = self:getValidFarmTitle(farmId)

            g_gui:showYesNoDialog({
                title = farmName,
                text = message,
                yesText = g_i18n:getText("action_sendOffTrain"),
                noText = g_i18n:getText("button_cancel"),
                dialogType = DialogElement.TYPE_INFO,
                callback = self.onConfirmStartRequest,
                target = self
            })
        else
            if message ~= "" then
                g_gui:showInfoDialog({
                    text = message,
                    dialogType = dialogType
                })
            end
        end
    else
        g_gui:showInfoDialog({
            text = g_i18n:getText("ui_noFarm"),
            dialogType = DialogElement.TYPE_WARNING
        })
    end
end

function SellingStationTrain.getTableSize(t)
    if t ~= nil then
        return ListUtil.size(t)
    end

    return 0
end


SellingStationTrainActivatable = {}

local SellingStationTrainActivatable_mt = Class(SellingStationTrainActivatable)

function SellingStationTrainActivatable:new(owner, interactionTrigger)
    local self = {}
    setmetatable(self, SellingStationTrainActivatable_mt)

    self.owner = owner
    self.interactionTrigger = interactionTrigger

    if interactionTrigger ~= nil then
        addTrigger(interactionTrigger, "interactionTriggerCallback", self)
    end

    self.isEnabled = true
    self.objectActivated = false

    self.activateText = g_i18n:getText("input_INTERACT")

    return self
end

function SellingStationTrainActivatable:delete()
    self.isEnabled = false

    if self.objectActivated then
        g_currentMission:removeActivatableObject(self)
        self.objectActivated = false
    end

    if self.interactionTrigger ~= nil then
        removeTrigger(self.interactionTrigger)
        self.interactionTrigger = nil
    end
end

function SellingStationTrainActivatable:getIsActivatable()
    if self.isEnabled and self.owner ~= nil then
        return g_currentMission.controlPlayer
    end

    return false
end

function SellingStationTrainActivatable:drawActivate()
    return
end

function SellingStationTrainActivatable:onActivateObject()
    if self.isEnabled and self.owner ~= nil then
        self.owner:requestToStartTrain(g_currentMission.player.farmId)
    end
end

function SellingStationTrainActivatable:shouldRemoveActivatable()
    return false
end

function SellingStationTrainActivatable:interactionTriggerCallback(triggerId, otherId, onEnter, onLeave, onStay)
    if self.isEnabled and self.owner ~= nil then
        if onEnter or onLeave then
            if g_currentMission.player ~= nil and otherId == g_currentMission.player.rootNode then
                if onEnter then
                    if not self.objectActivated then
                        g_currentMission:addActivatableObject(self)
                        self.objectActivated = true
                    end
                else
                    if self.objectActivated then
                        g_currentMission:removeActivatableObject(self)
                        self.objectActivated = false
                    end
                end
            end
        end
    end
end


SellingStationTrainFillLevelEvent = {}

local SellingStationTrainFillLevelEvent_mt = Class(SellingStationTrainFillLevelEvent, Event)
InitEventClass(SellingStationTrainFillLevelEvent, "SellingStationTrainFillLevelEvent")

function SellingStationTrainFillLevelEvent:emptyNew()
    local self = Event:new(SellingStationTrainFillLevelEvent_mt)

    return self
end

function SellingStationTrainFillLevelEvent:new(object)
    local self = SellingStationTrainFillLevelEvent:emptyNew()

    self.object = object

    return self
end

function SellingStationTrainFillLevelEvent:readStream(streamId, connection)
    local object = NetworkUtil.readNodeObject(streamId)
    local currentFarmId = g_currentMission:getFarmId()

    for farmId = 1, object.maxNumFarms do
        local numFillTypes = streamReadUInt8(streamId)

        if numFillTypes > 0 then
            for i = 1, numFillTypes do
                local fillLevel, lastAdded = nil, 0
                local fillTypeIndex = streamReadUIntN(streamId, FillTypeManager.SEND_NUM_BITS)

                if streamReadBool(streamId) then
                    fillLevel = streamReadFloat32(streamId)
                end

                if fillLevel ~= nil and farmId == currentFarmId then
                    lastAdded = fillLevel - object:getFillLevel(farmId, fillTypeIndex)
                end

                object:setFillLevel(farmId, fillTypeIndex, fillLevel, lastAdded, nil, lastAdded > 0)
            end
        end
    end
end

function SellingStationTrainFillLevelEvent:writeStream(streamId, connection)
    NetworkUtil.writeNodeObject(streamId, self.object)

    for farmId = 1, self.object.maxNumFarms do
        local fillLevels = self.object.fillLevels[farmId]
        local numFillTypes = SellingStationTrain.getTableSize(fillLevels)

        streamWriteUInt8(streamId, numFillTypes)

        if numFillTypes > 0 then
            for fillTypeIndex, fillLevel in pairs (fillLevels) do
                streamWriteUIntN(streamId, fillTypeIndex, FillTypeManager.SEND_NUM_BITS)

                if streamWriteBool(streamId, fillLevel > 0) then
                    streamWriteFloat32(streamId, fillLevel)
                end
            end
        end
    end
end

function SellingStationTrainFillLevelEvent:run(connection)
    if not connection:getIsServer() then
        print("  Info: Even 'SellingStationTrainFillLevelEvent' is a Server > Client only event!")
    end
end


SellingStationTrainStartRequestEvent = {}

local SellingStationTrainRequestEvent_mt = Class(SellingStationTrainStartRequestEvent, Event)
InitEventClass(SellingStationTrainStartRequestEvent, "SellingStationTrainStartRequestEvent")

function SellingStationTrainStartRequestEvent:emptyNew()
    local self = Event:new(SellingStationTrainRequestEvent_mt)

    return self
end

function SellingStationTrainStartRequestEvent:new(object, userId)
    local self = SellingStationTrainStartRequestEvent:emptyNew()

    self.object = object
    self.userId = userId

    return self
end

function SellingStationTrainStartRequestEvent:readStream(streamId, connection)
    self.object = NetworkUtil.readNodeObject(streamId)
    self.userId = streamReadInt8(streamId)

    self:run(connection)
end

function SellingStationTrainStartRequestEvent:writeStream(streamId, connection)
    NetworkUtil.writeNodeObject(streamId, self.object)
    streamWriteInt8(streamId, self.userId)
end

function SellingStationTrainStartRequestEvent:run(connection)
    if not connection:getIsServer() then
        if self.object ~= nil then
            if self.object:getIsTrainReadyForTravel() then
                g_server:broadcastEvent(SellingStationTrainStateEvent:new(self.object, self.userId, true), nil, nil, self.object)
                self.object:setTravelDirection(SellingStationTrain.DIRECTION_FOWARD, true, self.userId, false)
            else
                connection:sendEvent(SellingStationTrainStateEvent:new(self.object, self.userId, false))
            end
        end
    end
end

function SellingStationTrainStartRequestEvent.sendEvent(object, userId)
    if object ~= nil and userId ~= nil then
        if g_server ~= nil then
            g_server:broadcastEvent(SellingStationTrainStateEvent:new(object, userId, true), nil, nil, object)
            object:setTravelDirection(SellingStationTrain.DIRECTION_FOWARD, true, userId, false)
        else
            g_client:getServerConnection():sendEvent(SellingStationTrainStartRequestEvent:new(object, userId))
        end
    end
end


SellingStationTrainStateEvent = {}

local SellingStationTrainStateEvent_mt = Class(SellingStationTrainStateEvent, Event)
InitEventClass(SellingStationTrainStateEvent, "SellingStationTrainStateEvent")

function SellingStationTrainStateEvent:emptyNew()
    local self = Event:new(SellingStationTrainStateEvent_mt)

    return self
end

function SellingStationTrainStateEvent:new(object, userId, startTrain)
    local self = SellingStationTrainStateEvent:emptyNew()

    self.object = object
    self.userId = userId
    self.startTrain = startTrain

    return self
end

function SellingStationTrainStateEvent:readStream(streamId, connection)
    self.object = NetworkUtil.readNodeObject(streamId)
    self.userId = streamReadInt8(streamId)
    self.startTrain = streamReadBool(streamId)

    self:run(connection)
end

function SellingStationTrainStateEvent:writeStream(streamId, connection)
    NetworkUtil.writeNodeObject(streamId, self.object)
    streamWriteInt8(streamId, self.userId)
    streamWriteBool(streamId, self.startTrain)
end

function SellingStationTrainStateEvent:run(connection)
    if connection:getIsServer() then
        if self.startTrain then
            if self.object ~= nil then
                self.object:setTravelDirection(SellingStationTrain.DIRECTION_FOWARD, true, self.userId, false)
            end
        else
            if g_currentMission.playerUserId == self.userId then
                g_gui:showInfoDialog({
                    text = g_i18n:getText("warning_actionNotAllowedNow"),
                    dialogType = DialogElement.TYPE_WARNING
                })
            end
        end
    end
end

SellingStationTrainSetDirectionEvent = {}

local SellingStationTrainSetDirectionEvent_mt = Class(SellingStationTrainSetDirectionEvent, Event)
InitEventClass(SellingStationTrainSetDirectionEvent, "SellingStationTrainSetDirectionEvent")

function SellingStationTrainSetDirectionEvent:emptyNew()
    local self = Event:new(SellingStationTrainSetDirectionEvent_mt)

    return self
end

function SellingStationTrainSetDirectionEvent:new(object, direction)
    local self = SellingStationTrainSetDirectionEvent:emptyNew()
    self.object = object
    self.direction = MathUtil.sign(direction)

    return self
end

function SellingStationTrainSetDirectionEvent:readStream(streamId, connection)
    self.object = NetworkUtil.readNodeObject(streamId)
    self.direction = streamReadUIntN(streamId, 2) - 1

    self:run(connection)
end

function SellingStationTrainSetDirectionEvent:writeStream(streamId, connection)
    NetworkUtil.writeNodeObject(streamId, self.object)
    streamWriteUIntN(streamId, self.direction + 1, 2)
end

function SellingStationTrainSetDirectionEvent:run(connection)
    if connection:getIsServer() and self.object ~= nil then
        if self.direction == SellingStationTrain.DIRECTION_PARKED then
            self.object:updateSplineTime(self.object.trainLengthSplineTime)
        else
            self.object:updateSplineTime(self.direction)
        end

        self.object:setTravelDirection(self.direction, false, nil, false)
    end
end
