--[[
Interface: 1.7.1.0 b10490

Copyright (C) GtX (Andy), 2020

Author: GtX | Andy
Date: 14.09.2020
Version: 1.0.0.0

Contact:
https://forum.giants-software.com
https://github.com/GtX-Andy

Important:
Not to be added to any mods / maps or modified from its current release form.
No changes are to be made to this script without permission from GtX | Andy

Darf nicht zu Mods / Maps hinzugefügt oder von der aktuellen Release-Form geändert werden.
An diesem Skript dürfen ohne Genehmigung von GtX | Andy keine Änderungen vorgenommen werden
]]


local trainSystem
local loadError

local modName = g_currentModName
local modDirectory = g_currentModDirectory

local recomendedReleaseVersion = 2
local recomendedMapVersion = "1.1.0.0"

local function getMapVersionIsValid(directory)
    if directory ~= nil then
        local xmlFile = loadXMLFile("mapModDesc", directory .. "modDesc.xml")

        if xmlFile ~= 0 then
            local id = Utils.getNoNil(getXMLString(xmlFile, "modDesc.maps.map#id"), "")
            local releaseVersion = Utils.getNoNil(getXMLInt(xmlFile, "modDesc.maps.map#releaseVersion"), 1)

            delete(xmlFile)

            return id == "Sosnovka" and releaseVersion >= recomendedReleaseVersion
        end
    end

    return false
end

local function getEnvironmentIsValid(customEnv, author)
    -- Make sure it is the correct Sosnovka from modHub or addon will not install and may cause errors.
    return author == "GtX" and (customEnv == "FS19_Sosnovka" or customEnv == "FS19_Sosnovka_update")
end

local function getNodeIsValid(node)
    return node ~= nil and node ~= 0
end

local function updatePlaceableStoreItems(storeManager, mapCustomEnvironment)
    if storeManager == nil or mapCustomEnvironment == nil then
        return false
    end

    local name = "SOSNOVKA"
    local title = "Sosnovka"

    if g_i18n:hasText("title_sosnovka", mapCustomEnvironment) then
        title = g_i18n:getText("title_sosnovka", mapCustomEnvironment)
    end

    storeManager:addCategory(name, title, "store/category_sosnovka.dds", "PLACEABLE", modDirectory)

    if storeManager.categories[name] ~= nil then
        for _, storeItem in ipairs(storeManager.items) do
            if storeItem.customEnvironment == mapCustomEnvironment then
                local brand = g_brandManager:getBrandByIndex(storeItem.brandIndex)

                if (brand ~= nil and brand.name == name) and storeItem.categoryName ~= "MISC" then
                    storeItem.categoryName = name
                end
            end
        end
    end

    return true
end

local function bgaLoad(bga, superFunc, id, xmlFile, key, customEnvironment)
    local loaded = superFunc(bga, id, xmlFile, key, customEnvironment)

    if loaded and getEnvironmentIsValid(customEnvironment, "GtX") then
        local digestateStorage = bga.digestateSilo.storages[1]

        if digestateStorage ~= nil and digestateStorage.rootNode ~= nil then
            local fillPlaneNode = getChild(digestateStorage.rootNode, "liquidManure")

            if fillPlaneNode ~= nil then
                local x, y, z = getTranslation(fillPlaneNode)

                digestateStorage.fillPlane = FillPlane:new()

                digestateStorage.fillPlane.loaded = true
                digestateStorage.fillPlane.colorChange = false

                digestateStorage.fillPlane.moveMinY = 0.3
                digestateStorage.fillPlane.moveMaxY = y

                digestateStorage.fillPlane.visibility = false
                digestateStorage.fillPlane.node = fillPlaneNode

                setTranslation(fillPlaneNode, x, digestateStorage.fillPlane.moveMinY, z)
                setVisibility(fillPlaneNode, false)

                digestateStorage.setFillLevel = Utils.overwrittenFunction(digestateStorage.setFillLevel, function(storage, superFunc, fillLevel, fillType)
                    superFunc(storage, fillLevel, fillType)

                    local newFillLevel = storage:getFillLevel(fillType)

                    if storage.fillPlane ~= nil and storage.fillPlane:setState(MathUtil.clamp(newFillLevel / storage.capacityPerFillType, 0, 1)) then
                        local visibility = newFillLevel > 0.0001

                        if storage.fillPlane.visibility ~= visibility then
                            storage.fillPlane.visibility = visibility
                            setVisibility(storage.fillPlane.node, visibility)
                        end
                    end

                    if storage.display ~= nil then
                        I3DUtil.setNumberShaderByValue(storage.display, math.max(0, newFillLevel), 0, true)
                    end
                end)
            end

            local display = I3DUtil.indexToObject(bga.nodeId, "0|1")

            if display ~= nil and getName(display) == "display" then
                local digestateStorageDisplay = clone(display, true, false, false)

                if digestateStorageDisplay ~= nil then
                    digestateStorage.display = I3DUtil.indexToObject(digestateStorageDisplay, "0|2")

                    if digestateStorage.display ~= nil and getName(digestateStorage.display) == "numbers" then
                        setName(display, "bunkerDisplay")
                        setName(digestateStorageDisplay, "digestateDisplay")

                        setTranslation(digestateStorageDisplay, 32.3, 0.105, -4.838)
                        setRotation(digestateStorageDisplay, 0, math.rad(-90), 0)

                        I3DUtil.setNumberShaderByValue(digestateStorage.display, 0, 0, true)
                    else
                        delete(digestateStorage.display)
                    end
                end
            end
        end
    end

    return loaded
end

local function woodSellStationCalculateBaseValue(woodSellStation, superFunc, objectId)
    if woodSellStation.priceMultiplier == nil then
        woodSellStation.priceMultiplier = tonumber(Utils.getNoNil(getUserAttribute(woodSellStation.woodSellTrigger, "priceMultiplier"), 1))
    end

    return superFunc(woodSellStation, objectId) * woodSellStation.priceMultiplier
end

local function addLoadError(message)
    if g_dedicatedServerInfo == nil and message ~= nil then
        loadError = {
            message = message,
            modName = modName,
            startUpdateTime = 1600,

            update = function(self, dt)
                self.startUpdateTime = self.startUpdateTime - dt

                if self.startUpdateTime < 0 then
                    if not g_gui:getIsGuiVisible() then
                        local hud = g_currentMission.hud

                        if hud ~= nil and not hud:isInGameMessageVisible() and (hud.popupMessage == nil or #hud.popupMessage.pendingMessages == 0) then
                            g_gui:showYesNoDialog({
                                title = "",
                                text = message,
                                dialogType = DialogElement.TYPE_LOADING,
                                callback = self.openModHubLink,
                                target = nil,
                                yesText = g_i18n:getText("button_ok"),
                                noText = g_i18n:getText("button_modHubDownload")
                            })
                        end
                    end
                end
            end,

            openModHubLink = function(ignore)
                if ignore == false then
                    local language = g_languageShort
                    local link = "mods.php?lang=en&country=be&title=fs2019&filter=org&org_id=129652&page=0"
                    if language == "de" or language == "fr" then
                        link = "mods.php?lang=" .. language .. "&country=be&title=fs2019&filter=org&org_id=129652&page=0"
                    end

                    openWebFile(link, "")
                end

                removeModEventListener(loadError)
                loadError = nil
            end
        }

        addModEventListener(loadError)
    end
end

local function loadMapFinished(mission, node)
    if not mission.cancelLoading and g_sosnovkaTrainSystem == nil then
        local map = g_modManager:getModByName(mission.loadingMapModName)

        if map ~= nil and getEnvironmentIsValid(map.modName, map.author) then
            if not getMapVersionIsValid(map.modDir) then
                addLoadError(string.format(g_i18n:getText("info_updateMap", modName), map.version, modName, recomendedMapVersion))
            end

            -- Add custom train system to map.
            if getNodeIsValid(node) then
                local gameplayNode = getChild(node, "gameplay")

                if getNodeIsValid(gameplayNode) then
                    local addonNode = getChild(gameplayNode, "sosnovkaAddon")

                    if getNodeIsValid(addonNode) and getUserAttribute(addonNode, "sosnovkaAddon") then
                        trainSystem = SosnovkaTrainSystem.init(addonNode, mission)
                        getfenv(0)["g_sosnovkaTrainSystem"] = trainSystem
                    else
                        g_logManager:info(" [SosnovkaAddon]  Failed to find addon node 'sosnovkaAddon' or attribute 'sosnovkaAddon' is missing or set to 'false'!")
                    end
                end
            end

            -- Move all placeables to the Sosnovka category if it was created or exists already.
            updatePlaceableStoreItems(g_storeManager, mission.loadingMapModName)

            -- Add moving fillplane and digital display to liquid manure storage.
            Bga.load = Utils.overwrittenFunction(Bga.load, bgaLoad)

            -- Add missing attribute 'priceMultiplier' to the 'woodSellTrigger' node as it was in previous games.
            if not WoodSellStationPlaceable.HAS_PRICE_MULTIPLIER then
                WoodSellStationPlaceable.HAS_PRICE_MULTIPLIER = true

                WoodSellStationPlaceable.calculateBaseValue = Utils.overwrittenFunction(WoodSellStationPlaceable.calculateBaseValue, woodSellStationCalculateBaseValue)
            end

            -- Multiplayer fix when buying land a silo is positioned on. It seems this is not updated for the clients when land is bought and sold.
            SiloPlaceable.setOwnerFarmId = Utils.appendedFunction(SiloPlaceable.setOwnerFarmId, function (placeable, farmId, noEventSend)
                if not placeable.isServer and placeable.storages ~= nil and not placeable.storagePerFarm then
                    for _, storage in ipairs(placeable.storages) do
                        storage:setOwnerFarmId(farmId, true)
                    end
                end
            end)

            -- Multiplayer fix for placeable delete, it does not remove the listener for clients even though it is added at finalizePlacement.
            -- Only need to fix for the 2 Buying Stations as I do this at the main farm. So no need to prepend the Placeable.
            BuyingStationPlaceable.delete = Utils.prependedFunction(BuyingStationPlaceable.delete, function (placeable)
                if placeable.boughtWithFarmland and not placeable.isServer then
                    g_farmlandManager:removeStateChangeListener(placeable)
                end
            end)
        else
            addLoadError(string.format(g_i18n:getText("warning_invalidMapLoaded", modName), modName, "GtX"))
            g_logManager:info(" [SosnovkaAddon]  Invalid map, addon failed to load!")
        end
    end
end

local function unload(mission)
    if trainSystem ~= nil then
        if trainSystem.delete ~= nil then
            trainSystem:delete(mission)
        end

        if trainSystem == g_sosnovkaTrainSystem then
            getfenv(0)["g_sosnovkaTrainSystem"] = nil
        end
    end

    trainSystem = nil
    loadError = nil
end

local function init()
    if g_sosnovkaTrainSystem == nil then
        source(modDirectory .. "scripts/SosnovkaTrainSystem.lua")
        source(modDirectory .. "scripts/SellingStationTrain.lua")
        source(modDirectory .. "scripts/RailroadBarrier.lua")

        FSBaseMission.loadMapFinished = Utils.prependedFunction(FSBaseMission.loadMapFinished, loadMapFinished)
        FSBaseMission.delete = Utils.appendedFunction(FSBaseMission.delete, unload)
    else
        g_logManager:error(" [SosnovkaAddon]  Failed to load, train system is already loaded by '%s'!", tostring(g_sosnovkaTrainSystem.customEnvironment))
    end
end

init()
