--[[
Interface: 1.7.1.0 b10490

Copyright (C) GtX (Andy), 2020

Author: GtX | Andy
Date: 14.09.2020
Version: 1.0.0.0

Contact:
https://forum.giants-software.com
https://github.com/GtX-Andy

Important:
Not to be added to any mods / maps or modified from its current release form.
No changes are to be made to this script without permission from GtX | Andy

Darf nicht zu Mods / Maps hinzugefügt oder von der aktuellen Release-Form geändert werden.
An diesem Skript dürfen ohne Genehmigung von GtX | Andy keine Änderungen vorgenommen werden
]]


SosnovkaTrainSystem = {}

local trainSystem

local modName = g_currentModName
local modDirectory = g_currentModDirectory

local SosnovkaTrainSystem_mt = Class(SosnovkaTrainSystem)

function SosnovkaTrainSystem.init(addonNode, mission)
    if g_sosnovkaTrainSystem == nil and trainSystem == nil and mission ~= nil then
        trainSystem = SosnovkaTrainSystem:new(mission:getIsServer(), mission:getIsClient())

        if trainSystem ~= nil then
            if trainSystem:load(addonNode, modDirectory, modName, mission) then

                if g_registeredConsoleCommands["stsToggleTrainSystemDebug"] == nil then
                    local missionInfo = mission.missionInfo

                    if missionInfo ~= nil and missionInfo.playerStyle ~= nil and missionInfo.playerStyle.playerName == "GtX" then
                        addConsoleCommand("stsToggleTrainSystemDebug", "Toggles the debug mode for the complete train system", "consoleCommandToggleDebug", trainSystem)
                        trainSystem.consoleCommandLoaded = true
                    end
                end

                return trainSystem
            else
                trainSystem:delete(mission)
                g_logManager:error(" [SosnovkaTrainSystem]  Failed to create train system!")
            end
        end
    end

    return nil
end

function SosnovkaTrainSystem:new(isServer, isClient, customMt)
    local self = setmetatable({}, customMt or SosnovkaTrainSystem_mt)

    self.isServer = isServer
    self.isClient = isClient

    self.addonNode = 0
    self.barriersNode = 0

    self.decoNode = 0
    self.trainSystemNode = 0

    self.trainSystems = {}
    self.railroadBarriers = {}

    self.xmlFilenameToTrainSystemData = {}
    self.gameLoaded = false

    self.debugActive = false
    self.consoleCommandLoaded = false

    return self
end

function SosnovkaTrainSystem:load(addonNode, baseDirectory, customEnvironment, mission)
    if mission == nil or mission.gameStarted then
        g_logManager:error(" [SosnovkaTrainSystem]  Load failed, game has already been started!")

        return false
    end

    self.addonNode = addonNode

    self.decoNode = I3DUtil.indexToObject(addonNode, getUserAttribute(addonNode, "decoTrains"))
    self.trainSystemNode = I3DUtil.indexToObject(addonNode, Utils.getNoNil(getUserAttribute(addonNode, "sellingStationTrains"), getUserAttribute(addonNode, "buyStationTrains")))

    self.mapModName = mission.loadingMapModName
    self.mapBaseDirectory = mission.loadingMapBaseDirectory

    self.baseDirectory = baseDirectory
    self.customEnvironment = customEnvironment

    self.i3dFilename = Utils.getFilename("trainSystem/sosnovkaTrainSystem.i3d", baseDirectory)

    local success, numberTrains, numberBarriers = self:loadTrainSystem()

    if success then
        if numberTrains > 0 then
            WoodSellStationPlaceable.load = Utils.overwrittenFunction(WoodSellStationPlaceable.load, SosnovkaTrainSystem.inj_woodSellStationPlaceable_load)
            SellingStationPlaceable.load = Utils.overwrittenFunction(SellingStationPlaceable.load, SosnovkaTrainSystem.inj_sellingStationPlaceable_load)

            WoodSellStationPlaceable.loadFromXMLFile = Utils.overwrittenFunction(WoodSellStationPlaceable.loadFromXMLFile, SosnovkaTrainSystem.inj_woodSellStationPlaceable_loadFromXMLFile)
            SellingStationPlaceable.loadFromXMLFile = Utils.overwrittenFunction(SellingStationPlaceable.loadFromXMLFile, SosnovkaTrainSystem.inj_sellingStationPlaceable_loadFromXMLFile)

            WoodSellStationPlaceable.readStream = Utils.appendedFunction(WoodSellStationPlaceable.readStream, SosnovkaTrainSystem.inj_woodSellStationPlaceable_readStream)
            SellingStationPlaceable.readStream = Utils.appendedFunction(SellingStationPlaceable.readStream, SosnovkaTrainSystem.inj_sellingStationPlaceable_readStream)

            WoodSellStationPlaceable.writeStream = Utils.appendedFunction(WoodSellStationPlaceable.writeStream, SosnovkaTrainSystem.inj_woodSellStationPlaceable_writeStream)
            SellingStationPlaceable.writeStream = Utils.appendedFunction(SellingStationPlaceable.writeStream, SosnovkaTrainSystem.inj_sellingStationPlaceable_writeStream)

            g_logManager:info(" [SosnovkaTrainSystem]  '%d' train spline(s) have been added.", numberTrains)
        end

        if numberBarriers > 0 then
            g_logManager:info(" [SosnovkaTrainSystem]  '%d' railroad barrier(s) have been added.", numberBarriers)
        end

        local numberDecoObjects = self:removeDecoObjects()

        if numberDecoObjects > 0 then
            g_logManager:info(" [SosnovkaTrainSystem]  '%d' train decoration object(s) have been removed.", numberDecoObjects)
        end

        mission:addLoadFinishedListener(self)

        return true
    end

    return false
end

function SosnovkaTrainSystem:delete(mission)
    if self.consoleCommandLoaded then
        removeConsoleCommand("stsToggleTrainSystemDebug")
        self.consoleCommandLoaded = false
    end

    self.trainSystems = {}
    self.railroadBarriers = {}

    self.gameLoaded = false
    trainSystem = nil
end

function SosnovkaTrainSystem:loadTrainSystem()
    if not self.gameLoaded then
        local numberTrainsAdded, numberBarriersAdded = 0, 0

        if self.trainSystemNode ~= nil then
            local i3dRoot = loadI3DFile(self.i3dFilename, false, false, true)

            if i3dRoot ~= nil or i3dRoot ~= 0 then
                local childId = getChild(i3dRoot, "sellingStationTrains")

                if childId ~= nil and childId ~= 0 then
                    for i = 0, getNumOfChildren(self.trainSystemNode) - 1 do
                        local linkNode = getChildAt(self.trainSystemNode, i)
                        local linkNodeName = getName(linkNode)

                        for j = 0, getNumOfChildren(childId) - 1 do
                            local node = getChildAt(childId, j)

                            if getName(node) == linkNodeName then
                                local filename = Utils.getNoNil(getUserAttribute(node, "placeableFilename"), "")
                                local xmlFilename = getUserAttribute(node, "xmlFilename")

                                if filename ~= "" and xmlFilename ~= nil then
                                    filename = NetworkUtil.convertFromNetworkFilename(filename):upper()

                                    self.xmlFilenameToTrainSystemData[filename] = {
                                        xmlFilename = Utils.getFilename(xmlFilename, self.baseDirectory),
                                        node = linkNode,
                                        name = linkNodeName
                                    }

                                    link(linkNode, node)
                                    addToPhysics(node)

                                    numberTrainsAdded = numberTrainsAdded + 1
                                end

                                break
                            end
                        end
                    end
                end

                if RailroadBarrier ~= nil  then
                    childId = getChild(i3dRoot, "railroadBarriers")

                    if childId ~= nil and childId ~= 0 then
                        link(self.addonNode, childId)
                        addToPhysics(childId)

                        self.barriersNode = childId

                        for i = 0, getNumOfChildren(self.barriersNode) - 1 do
                            local railroadBarrier = getChildAt(self.barriersNode, i)

                            if RailroadBarrier:createFromNode(railroadBarrier, self.baseDirectory, self.customEnvironment) then
                                numberBarriersAdded = numberBarriersAdded + 1
                                self.railroadBarriers[numberBarriersAdded] = railroadBarrier
                            end
                        end
                    end
                end

                delete(i3dRoot)
            else
                g_logManager:error(" [SosnovkaTrainSystem]  Failed to load i3d file '" .. filename .. "'!")
            end
        else
            g_logManager:error(" [SosnovkaTrainSystem]  Failed to find valid addon link node!")
        end

        return true, numberTrainsAdded, numberBarriersAdded
    end

    return false, 0, 0
end

function SosnovkaTrainSystem:removeDecoObjects()
    local numberRemoved = 0

    if not self.gameLoaded and self.decoNode ~= nil then
        for i = getNumOfChildren(self.decoNode) - 1, 0, -1 do
            delete(getChildAt(self.decoNode, i))
            numberRemoved = numberRemoved + 1
        end
    end

    return numberRemoved
end

function SosnovkaTrainSystem:onLoadFinished()
    self.gameLoaded = true
    self.debugActive = false
end

function SosnovkaTrainSystem:getTrainsSystemData(xmlFilename)
    if xmlFilename ~= nil then
        return self.xmlFilenameToTrainSystemData[xmlFilename:upper()]
    end

    return nil
end

function SosnovkaTrainSystem:getTrainSystemLoaded()
    return self:getNumberOfTrainsLoaded() > 0
end

function SosnovkaTrainSystem:getNumberOfTrainsLoaded()
    return #self.trainSystems
end

function SosnovkaTrainSystem:consoleCommandToggleDebug()
    self.debugActive = not self.debugActive

    for _, sellingStationTrain in ipairs (self.trainSystems) do
        sellingStationTrain.debugActive = self.debugActive

        local debugListener

        if sellingStationTrain.splinePositionListenersActive then
            for _, listener in ipairs(sellingStationTrain.splinePositionUpdateListeners) do
                if listener.isDebugListener then
                    debugListener = listener
                    break
                end
            end
        end

        if sellingStationTrain.debugActive then
            if debugListener == nil then
                sellingStationTrain:setPositionListeners()
                sellingStationTrain:setNeedsUpdate()
            end
        else
            if debugListener ~= nil and debugListener.onTrainDelete ~= nil then
                debugListener:onTrainDelete(sellingStationTrain)
            end
        end
    end

    if RailroadBarrier ~= nil then
        RailroadBarrier.DEBUG_ACTIVE = self.debugActive
    end

    return "SosnovkaTrainSystemDebug = " .. tostring(self.debugActive)
end

function SosnovkaTrainSystem.inj_woodSellStationPlaceable_load(placeable, superFunc, xmlFilename, x, y, z, rx, ry, rz, initRandom)
    if superFunc(placeable, xmlFilename, x, y, z, rx, ry, rz, initRandom) then
        if trainSystem ~= nil and not trainSystem.gameLoaded then
            local trainSystemData = trainSystem:getTrainsSystemData(xmlFilename)

            if trainSystemData ~= nil then
                local xmlFile = loadXMLFile("TempXML", trainSystemData.xmlFilename)

                if xmlFile ~= nil or xmlFile ~= 0 then
                    local instance = SellingStationTrain:new(placeable.isServer, placeable.isClient)

                    if instance ~= nil and instance:load(trainSystemData.node, placeable, xmlFile, "sellingStationTrain", modName, modDirectory, trainSystemData.xmlFilename) then
                        instance:register(true)

                        placeable.sellingStationTrain = instance
                        placeable.activateText = g_i18n:getText("sellingStationTrain_loadTrain", modName)

                        placeable.delete = Utils.prependedFunction(placeable.delete, function(woodSellStation)
                            local sellingStationTrain = woodSellStation.sellingStationTrain

                            if sellingStationTrain ~= nil then
                                sellingStationTrain:delete()

                                if sellingStationTrain.nodeId ~= nil and sellingStationTrain.nodeId ~= 0 then
                                    if entityExists(sellingStationTrain.nodeId) then
                                        delete(sellingStationTrain.nodeId)
                                        sellingStationTrain.nodeId = nil
                                    end
                                end
                            end
                        end)

                        placeable.saveToXMLFile = Utils.appendedFunction(placeable.saveToXMLFile, function(woodSellStation, xmlFile, key, usedModNames)
                            if woodSellStation.sellingStationTrain ~= nil then
                                woodSellStation.sellingStationTrain:saveToXMLFile(xmlFile, key .. ".sellingStationTrain", usedModNames)
                            end
                        end)

                        placeable.onActivateObject = Utils.overwrittenFunction(placeable.onActivateObject, function(woodSellStation, superFunc)
                            if woodSellStation.sellingStationTrain ~= nil then
                                local farmId = g_currentMission:getFarmId()

                                g_currentMission:addActivatableObject(woodSellStation)

                                if woodSellStation.sellingStationTrain:getIsTrainReadyForDelivery(woodSellStation, FillType.WOODCHIPS, farmId, true) then
                                    woodSellStation.objectActivated = true
                                    woodSellStation:sellWood(farmId)
                                else
                                    g_gui:showInfoDialog({
                                        text = g_i18n:getText("sellingStationTrain_noTrainAvailable", modName),
                                        dialogType = DialogElement.TYPE_LOADING
                                    })
                                end
                            end

                            return superFunc(woodSellStation)
                        end)

                        placeable.sellWood = Utils.overwrittenFunction(placeable.sellWood, function(woodSellStation, superFunc, farmId)
                            if woodSellStation.sellingStationTrain == nil then
                                superFunc(woodSellStation, farmId)
                            end

                            if g_currentMission:getIsServer() then
                                local numLogs = 0

                                local price = 0
                                local volume = 0

                                local calculatedValue = 0
                                local difficultyMultiplier = g_currentMission.missionInfo.sellPriceMultiplier or 1

                                for _, nodeId in pairs(woodSellStation.woodInTrigger) do
                                    if entityExists(nodeId) then
                                        calculatedValue = woodSellStation:calculateBaseValue(nodeId) * difficultyMultiplier

                                        price = price + calculatedValue
                                        volume = volume + getVolume(nodeId)

                                        delete(nodeId)
                                        numLogs = numLogs + 1
                                    end

                                    woodSellStation.woodInTrigger[nodeId] = nil
                                end

                                if numLogs > 0 then
                                    -- Not sure what uses this but sends price like this when testing so send here.
                                    for _, listener in pairs(woodSellStation.updateEventListeners) do
                                        listener:onWoodSellingUpdateEvent(woodSellStation, price)
                                    end

                                    woodSellStation.sellingStationTrain:addFillLevelToTrain(farmId, FillType.WOODCHIPS, volume, price, numLogs)
                                end
                            else
                                g_client:getServerConnection():sendEvent(WoodSellEvent:new(woodSellStation, farmId))
                            end
                        end)

                        if trainSystem.trainSystems == nil then
                            trainSystem.trainSystems = {}
                        end

                        table.insert(trainSystem.trainSystems, instance)

                        local stationName = ""

                        if placeable.stationName ~= nil then
                            stationName = "'" .. tostring(placeable.stationName) .. "'"
                        end

                        g_logManager:info(" [SosnovkaTrainSystem]  Train system successfully added to Wood Sell Station %s.", stationName)
                    end

                    delete(xmlFile)
                else
                    g_logManager:error(" [SosnovkaTrainSystem]  Configuration file '%s' could not be loaded for node '%s'!", trainSystemData.xmlFilename, trainSystemData.name)
                end
            end
        end

        return true
    end

    return false
end

function SosnovkaTrainSystem.inj_sellingStationPlaceable_load(placeable, superFunc, xmlFilename, x, y, z, rx, ry, rz, initRandom)
    if superFunc(placeable, xmlFilename, x, y, z, rx, ry, rz, initRandom) then
        if trainSystem ~= nil and not trainSystem.gameLoaded and placeable.sellingStation ~= nil then
            local trainSystemData = trainSystem:getTrainsSystemData(xmlFilename)

            if trainSystemData ~= nil then
                local xmlFile = loadXMLFile("TempXML", trainSystemData.xmlFilename)

                if xmlFile ~= nil or xmlFile ~= 0 then
                    local instance = SellingStationTrain:new(placeable.isServer, placeable.isClient)

                    if instance ~= nil and instance:load(trainSystemData.node, placeable, xmlFile, "sellingStationTrain", modName, modDirectory, trainSystemData.xmlFilename) then
                        instance:register(true)

                        placeable.sellingStationTrain = instance

                        placeable.delete = Utils.prependedFunction(placeable.delete, function(sellingStationPlaceable)
                            local sellingStationTrain = sellingStationPlaceable.sellingStationTrain

                            if sellingStationTrain ~= nil then
                                sellingStationTrain:delete()

                                if sellingStationTrain.nodeId ~= nil and sellingStationTrain.nodeId ~= 0 then
                                    if entityExists(sellingStationTrain.nodeId) then
                                        delete(sellingStationTrain.nodeId)
                                        sellingStationTrain.nodeId = nil
                                    end
                                end
                            end
                        end)

                        placeable.saveToXMLFile = Utils.appendedFunction(placeable.saveToXMLFile, function(sellingStationPlaceable, xmlFile, key, usedModNames)
                            if sellingStationPlaceable.sellingStationTrain ~= nil then
                                sellingStationPlaceable.sellingStationTrain:saveToXMLFile(xmlFile, key .. ".sellingStationTrain", usedModNames)
                            end
                        end)

                        placeable.sellingStation.addFillLevelFromTool = Utils.overwrittenFunction(placeable.sellingStation.addFillLevelFromTool, function(sellingStation, superFunc, farmId, deltaFillLevel, fillType, fillInfo, toolType)
                            if sellingStation.owningPlaceable ~= nil and sellingStation.owningPlaceable.sellingStationTrain == nil then
                                return superFunc(sellingStation, farmId, deltaFillLevel, fillType, fillInfo, toolType)
                            end

                            if deltaFillLevel > 0 and sellingStation:getIsFillTypeAllowed(fillType) and sellingStation:getIsToolTypeAllowed(toolType) then
                                -- Check if this is used for mission first (Needs testing as undocumented!!).
                                if sellingStation.missions ~= nil then
                                    for _, mission in pairs(sellingStation.missions) do
                                        if mission.farmId == farmId and mission.fillType == fillType and mission.fillSold ~= nil then
                                            mission:fillSold(deltaFillLevel)

                                            return deltaFillLevel
                                        end
                                    end
                                end

                                if sellingStation.owningPlaceable.sellingStationTrain:getIsTrainReadyForTravel() then
                                    if deltaFillLevel == sellingStation.owningPlaceable.sellingStationTrain:addFillLevelToTrain(farmId, fillType, deltaFillLevel, 0, 0) then
                                        sellingStation:sellFillType(farmId, deltaFillLevel, fillType, toolType)

                                        local fillTypeDesc = g_fillTypeManager:getFillTypeByIndex(fillType)
                                        g_logManager:info("[SosnovkaTrainSystem] Something went wrong, failed to add '%d litres or %s' to train, selling instead!", deltaFillLevel, tostring(fillTypeDesc.title))
                                    end

                                    return deltaFillLevel
                                end
                            end

                            return 0
                        end)

                        placeable.sellingStation.getFreeCapacity = Utils.overwrittenFunction(placeable.sellingStation.getFreeCapacity, function(sellingStation, superFunc, fillTypeIndex, farmId)
                            local value = superFunc(sellingStation, fillTypeIndex, farmId)

                            if sellingStation.owningPlaceable ~= nil and sellingStation.owningPlaceable.sellingStationTrain ~= nil then
                                if not sellingStation.owningPlaceable.sellingStationTrain:getIsTrainReadyForDelivery(sellingStation, fillTypeIndex, farmId, false) then
                                    return 0
                                end
                            end

                            return value
                        end)

                        if trainSystem.trainSystems == nil then
                            trainSystem.trainSystems = {}
                        end

                        table.insert(trainSystem.trainSystems, instance)

                        local stationName = ""

                        if placeable.sellingStation.stationName ~= nil then
                            stationName = "'" .. tostring(placeable.sellingStation.stationName) .. "'"
                        end

                        g_logManager:info(" [SosnovkaTrainSystem]  Train system successfully added to Selling Station %s.", stationName)
                    end

                    delete(xmlFile)
                else
                    g_logManager:error(" [SosnovkaTrainSystem]  Configuration file '%s' could not be loaded for node '%s'!", trainSystemData.xmlFilename, trainSystemData.name)
                end
            end
        end

        return true
    end

    return false
end

function SosnovkaTrainSystem.inj_woodSellStationPlaceable_loadFromXMLFile(placeable, superFunc, xmlFile, key, resetVehicles)
    if superFunc(placeable, xmlFile, key, resetVehicles) then
        if placeable.sellingStationTrain ~= nil then
            placeable.sellingStationTrain:loadFromXMLFile(xmlFile, key .. ".sellingStationTrain", resetVehicles)
        end

        return true
    end

    return false
end

function SosnovkaTrainSystem.inj_sellingStationPlaceable_loadFromXMLFile(placeable, superFunc, xmlFile, key, resetVehicles)
    if superFunc(placeable, xmlFile, key, resetVehicles) then
        if placeable.sellingStationTrain ~= nil then
            placeable.sellingStationTrain:loadFromXMLFile(xmlFile, key .. ".sellingStationTrain", resetVehicles)
        end

        return true
    end

    return false
end

function SosnovkaTrainSystem.inj_sellingStationPlaceable_readStream(sellingStationPlaceable, streamId, connection)
    if connection:getIsServer() and sellingStationPlaceable.sellingStationTrain ~= nil then
        local sellingStationTrainId = NetworkUtil.readNodeObjectId(streamId)

        sellingStationPlaceable.sellingStationTrain:readStream(streamId, connection)
        g_client:finishRegisterObject(sellingStationPlaceable.sellingStationTrain, sellingStationTrainId)
    end
end

function SosnovkaTrainSystem.inj_sellingStationPlaceable_writeStream(sellingStationPlaceable, streamId, connection)
    if not connection:getIsServer() and sellingStationPlaceable.sellingStationTrain ~= nil then
        NetworkUtil.writeNodeObjectId(streamId, NetworkUtil.getObjectId(sellingStationPlaceable.sellingStationTrain))

        sellingStationPlaceable.sellingStationTrain:writeStream(streamId, connection)
        g_server:registerObjectInStream(connection, sellingStationPlaceable.sellingStationTrain)
    end
end

function SosnovkaTrainSystem.inj_woodSellStationPlaceable_readStream(woodSellStationPlaceable, streamId, connection)
    if connection:getIsServer() and woodSellStationPlaceable.sellingStationTrain ~= nil then
        local sellingStationTrainId = NetworkUtil.readNodeObjectId(streamId)

        woodSellStationPlaceable.sellingStationTrain:readStream(streamId, connection)
        g_client:finishRegisterObject(woodSellStationPlaceable.sellingStationTrain, sellingStationTrainId)
    end
end

function SosnovkaTrainSystem.inj_woodSellStationPlaceable_writeStream(woodSellStationPlaceable, streamId, connection)
    if not connection:getIsServer() and woodSellStationPlaceable.sellingStationTrain ~= nil then
        NetworkUtil.writeNodeObjectId(streamId, NetworkUtil.getObjectId(woodSellStationPlaceable.sellingStationTrain))

        woodSellStationPlaceable.sellingStationTrain:writeStream(streamId, connection)
        g_server:registerObjectInStream(connection, woodSellStationPlaceable.sellingStationTrain)
    end
end
